#ifndef lint
static char *rcsid = "$Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/add.c,v 0.95 91/09/02 11:32:45 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/add.c,v 0.95 91/09/02 11:32:45 emsrdsm Exp Locker: emsrdsm $ */

/*****************************************************************************

  add.c - Add entries

*****************************************************************************/

#include "types.h"
#include "util.h"
#include "read.h"
#include "add.h"
#include "request.h"
#include "ds_tasks.h"

#include "quipu/common.h"
#include "quipu/entry.h"
#include "quipu/add.h"
#include "quipu/name.h"
#include "quipu/util.h"
#include "quipu/dap2.h"
#include "quipu/attrvalue.h"
#include "quipu/photo.h"


/*
 *	Static procs..
 */

static errorList process_add();
static request_state process_add_resolve_request_result();


/*
 *	Externs..
 */

#ifndef NO_STATS
extern LLog *log_stat;
#endif


/*
 * - get_add_result() -
 *
 *
 */
addResult get_add_result(request_id)
     QCardinal request_id;
{
  requestRec request = _get_request_of_id(request_id);
  addRec addrec;
  addResult result;

  if (request == NULLReqRec)
    return NULLAddResult;
  
  addrec = request->ADD_REC;
  result = addrec->result;

  addrec->result = NULLAddResult;

  _request_complete(request_id);
  
  return result;
} /* get_add_result */




/*
 * - do_add() -
 *
 *
 */
errorList do_add(id_ptr, baseobject, ufn, rdn_type, add_attr_list)
     QCardinal *id_ptr;
     char *baseobject;
     char *ufn;
     AttributeType rdn_type;
     modifyAttr add_attr_list;
{
  addRec add_request;
  requestRec this_request;
  errorList errors;

  (void) _request_invoked(ADD_ENTRY, id_ptr);
  this_request = _get_request_of_id(*id_ptr);

  add_request = this_request->ADD_REC;

  add_request->request_id = *id_ptr;

  if (baseobject != NULLCP)
    {
      add_request->baseobject = copy_string(baseobject);
      add_request->ufn = NULLCP;
      add_request->rdn_type = NULLAttrT;
    }
  else
    {
      add_request->baseobject = NULLCP;
      add_request->ufn = copy_string(ufn);
      add_request->rdn_type = rdn_type;
    }

  add_request->add_attr_list = add_attr_list;

  if ((errors = process_add(*id_ptr,
			    baseobject,
			    ufn,
			    add_attr_list,
			    &add_request->task_id)) != NULLError)
    _request_complete(*id_ptr);

  return errors;
} /* do_add */

/*
 * - start_add() -
 *
 *
 */
static errorList process_add(request_id,
			     baseobject,
			     ufn,
			     add_attr_list,
			     task_id_ptr)
     QCardinal request_id;
     char *baseobject, *ufn;
     modifyAttr add_attr_list;
     int *task_id_ptr;
{
  int task_id;
  requestRec request_record = _get_request_of_id(request_id);
  addRec add_record = request_record->ADD_REC;

  struct ds_addentry_arg add_arg;
  struct DAPindication di;

  modifyAttr curr_attr;
  modifyValue curr_value;
  Attr_Sequence add_attrs = NULLATTR;
  AV_Sequence attrVal_seq = (AV_Sequence) NULL;
  AttributeValue attr_val;
  AttributeType curr_type = NULLAttrT;

  errorList errors = NULLError;
  char err_msg[LINESIZE];

  QE_error_code resolve_status;
  char *parent_name, *strval, *mapval;

  /* Quipu procedures */
  extern int avs_free(), as_free();
  extern char *char_unmap();

  if (baseobject == NULLCP)
    {
      for (parent_name = ufn;
	   !isnull(*parent_name) && *parent_name != ',';
	   parent_name++)
	if (*parent_name == '\"')
	  {
	    parent_name++;

	    while (!isnull(*parent_name) && *parent_name != '\"')
	      parent_name++;
	  }

      if (!isnull(*parent_name))
	parent_name++;

      while (!isnull(*parent_name) && isspace(*parent_name))
	parent_name++;

      if ((resolve_status = do_ufn_resolve(NULLEntryList,
					   parent_name,
					   might_not,
					   &add_record->sub_request_id))
	  != QERR_ok)
	{
          switch (resolve_status)
            {
            case QERR_bad_value_syntax:
              (void) sprintf(err_msg,
                             "*** Bad syntax in name `%.32s...' ! ***",
                             parent_name);
	      
	      add_error_to_list(&errors,
                                NULLCP,
                                QERR_bad_value_syntax,
                                (struct DSError *) NULL,
                                err_msg);

              return errors;

            case QERR_internal_limit_reached:
              add_error_to_list(&errors,
                                NULLCP,
                                QERR_bad_value_syntax,
                                (struct DSError *) NULL,
                        "*** Internal limit reached, can't perform! ***");
	      
	      return errors;

            default:
              add_error_to_list(&errors,
                                NULLCP,
                                QERR_internal,
                                (struct DSError *) NULL,
                                "*** Internal error, can't perform! ***");

              return errors;
            }
	}

      /*
       *	Resolve name initiated
       */
      
      tag_as_sub_request(add_record->sub_request_id, request_id);
      
      return NULLError;
    }

  /*
   *  Got a string DN, so attempt entry addition
   */

  for (curr_attr = add_attr_list;
       curr_attr != NULLModifyAttr;
       curr_attr = curr_attr->next)
    {
      attrVal_seq = NULLAV;
      
      curr_type = curr_attr->attr_type;

      for (curr_value = curr_attr->values;
	   curr_value != NULLModifyValue;
	   curr_value = curr_value->next)
	if ((curr_value->value != NULLCP && !isnull(*curr_value->value)) ||
	    (curr_value->new_value!=NULLCP && !isnull(*curr_value->new_value)))
	  {
	    AttributeValue str2file();

	    if (curr_value->status == deleted)
	      continue;
	    else if (curr_value->value!= NULLCP && !isnull(*curr_value->value))
	      strval = curr_value->value;
	    else
	      strval = curr_value->new_value;

	    if (curr_attr->attr_type->oa_syntax >= AV_WRITE_FILE)
	      attr_val = str2file(strval, curr_type);
	    else
	      {
		mapval = char_unmap(strval, curr_type);
		
		attr_val = str2AttrV(mapval, curr_type->oa_syntax);
		
		(void) free(mapval);
	      }

		
	    if (attr_val != NULLAttrV)
	      {
		if (attrVal_seq != NULLAV)
		  attrVal_seq = avs_merge(attrVal_seq,
					  avs_comp_new(attr_val));
		else
		  attrVal_seq = avs_comp_new(attr_val);
	      }
	    else
	      {
		if (curr_attr->attr_type->oa_syntax >= AV_WRITE_FILE)
		  (void) sprintf(err_msg,
		       "*** Bad file or filename in attribute value `%s'. ***",
				 strval);
		else
		  (void) sprintf(err_msg,
				 "*** Bad syntax in attribute value `%s'. ***",
				 strval);
		
		add_error_to_list(&errors,
				  NULLCP,
				  QERR_bad_value_syntax,
				  (struct DSError *) NULL,
				  err_msg);
		
		as_free(add_attrs);

		if (attrVal_seq != NULLAV)
		  avs_free(attrVal_seq);
		    
		return errors;
	      }
	  }
      
      if (attrVal_seq != NULLAV)
	{
	  if (add_attrs == NULLATTR)
	    add_attrs = as_comp_new(curr_type, attrVal_seq, NULLACL_INFO);
	  else
	    add_attrs = as_merge(as_comp_new(curr_type,
					     attrVal_seq,
					     NULLACL_INFO),
				 add_attrs);
	}
    }

  (void) get_default_service(&add_arg.ada_common);

  if ((add_arg.ada_object = str2dn(baseobject)) == NULLDN)
    {
      (void) sprintf(err_msg, "*** Bad entry name given `%s'", baseobject);
      add_error_to_list(&errors,
			NULLCP,
			QERR_bad_name,
			(struct DSError *) NULL,
			err_msg);

      as_free(add_attrs);

      return errors;
    }

  add_arg.ada_entry = add_attrs;

  
  if (_task_invoked(ADD_TASK, baseobject, request_id, &task_id) != QERR_ok)
    {
      add_error_to_list(&errors,
			NULLCP,
			QERR_internal_limit_reached,
			(struct DSError *) NULL,
			NULLCP);

      dn_free(add_arg.ada_object);
      as_free(add_attrs);

      return errors;
    }
  
  if (DapAddEntry(dsap_ad, task_id, &add_arg, &di, ROS_ASYNC) == NOTOK)
    {
      add_error_to_list(&errors,
			NULLCP,
			QERR_internal,
			(struct DSError *) NULL,
			NULLCP);

      dn_free(add_arg.ada_object);
      as_free(add_attrs);

      return errors;
    }
  
#ifndef NO_STATS
  LLOG(log_stat, LLOG_NOTICE, ("ADD ENTRY +%s, task %d", baseobject, task_id));
#endif

  dn_free(add_arg.ada_object);
  as_free(add_attrs);

  *task_id_ptr = task_id;

  return NULLError;
} /* start_add */

/*
 * - process_add_ds_result() -
 *
 *
 */
/* ARGSUSED */
request_state process_add_ds_result(request, task_id)
     requestRec request;
     int task_id;
{
  addRec addrec = request->ADD_REC;

  /* This shouldn't happen */
  if (_get_task_of_id(task_id) == NULLDsTask)
    return RQ_processing;

  addrec->result = add_result_alloc();
  
  addrec->result->was_successful = TRUE;
  addrec->result->errors = NULLError;

  addrec->result->baseobject = copy_string(addrec->baseobject);

  /* Remove invocation record for this task */
  _task_complete(task_id);

#ifndef NO_STATS
  LLOG(log_stat, LLOG_NOTICE,
       ("ADD ENTRY SUCCEEDED +%s, task %d",
	addrec->baseobject,
	task_id));
#endif

  return RQ_results_returned;
} /* process_add_ds_result */

/*
 * - process_add_ds_error() -
 *
 *
 */
request_state process_add_ds_error(request, task_id, ds_error)
     requestRec request;
     int task_id;
     struct DSError *ds_error;
{
  DsTask task_rec;
  addRec addrec = request->ADD_REC;
  QE_error_code error_type;

  /* This shouldn't happen */
  if ((task_rec = _get_task_of_id(task_id)) == NULLDsTask)
    return RQ_processing;

  /* Save error and log it */
  error_type = get_log_error_type(ds_error, task_id);
  add_error_to_request_rec(request,
                           task_rec->baseobject,
                           error_type,
                           ds_error,
                           NULLCP);

  addrec->result = add_result_alloc();
  addrec->result->was_successful = FALSE;

  addrec->result->errors = request->errors;
  request->errors = NULLError;

  addrec->result->baseobject = NULLCP;

  /* Remove invocation record for this task */
  _task_complete(task_id);

#ifndef NO_STATS
  LLOG(log_stat, LLOG_NOTICE,
       ("ADD ENTRY FAILED +%s, task %d",
	addrec->baseobject,
	task_id));
#endif

  return RQ_error_returned;
} /* process_add_ds_error */

/*
 * - process_add_resolve_request_result() -
 * Add only uses one sub request type-> name resolve
 *
 */
request_state process_add_resolve_result(sub_request_id)
     QCardinal sub_request_id;
{
  QCardinal request_id;
  requestRec request_record, sub_request_record;
  addRec add_record;
  
  ufnResults results;
  
  char err_msg[LINESIZE], baseobject[LINESIZE], rdn[LINESIZE];
  char *str;

  errorList errors;
  

  sub_request_record = _get_request_of_id(sub_request_id);
  request_id = sub_request_record->owner_request_id;

  request_record = _get_request_of_id(request_id);
  add_record = request_record->ADD_REC;


  results = get_ufn_results(sub_request_id);

  /*
   * Will have to change this if poor matches are to be queried
   */

  _request_complete(sub_request_id);


  /*
   *    Put errors into parent request record
   */

  for (errors = results->errors; errors != NULLError; errors = errors->next)
      if (errors->ds_message != NULLCP)
        {
          (void) sprintf(err_msg,
                         "While resolving parent of name `%.32s...'\n",
                         add_record->ufn);

          (void) strcat(err_msg, errors->ds_message);

          (void) free(errors->ds_message);
          errors->ds_message = copy_string(err_msg);
        }

  request_record->errors = results->errors;
  results->errors = NULLError;

  /*
   *    Has name resolution succeded (matches = 1) ?
   */

  if (results->match_status == GoodMatches && results->match_num == 1)
    {
      (void) strcpy(rdn, add_record->ufn);
      
      for (str = rdn; !isnull(*str) && *str != ','; str++)
	if (*str == '\"')
	  {
	    str++;

	    while (!isnull(*str) && *str != '\"')
	      str++;
	  }

      str--;

      while (isspace(*str) && str > rdn)
	str--;
      
      *++str = '\0';

      (void) sprintf(baseobject, "%s@%s=%s",
		     results->matches->string_dn,
		     add_record->rdn_type->oa_ot.ot_name,
		     rdn);

      add_record->baseobject = copy_string(baseobject);

      ufname_result_free(&results);

      if ((errors = process_add(add_record->request_id,
				baseobject,
				NULLCP,
				add_record->add_attr_list,
				&add_record->task_id))
	   != NULLError)
	{
	  errors->next = request_record->errors;
	  request_record->errors = NULLError;

	  add_record->result = add_result_alloc();
	  
	  add_record->result->errors = errors;
	  add_record->result->was_successful = FALSE;

	  add_record->result->baseobject = copy_string(baseobject);

	  return RQ_error_returned;
	}

      /*
       *	Add initiated
       */
      
      return RQ_processing;
    }
  else
    /*
     *  Name resolution failed
     */
    {
      if (results->match_num == 0)
	(void) sprintf(err_msg,
		       "*** Can't resolve parent name for `%.24s'! ***",
		       add_record->ufn);
      else
	(void) sprintf(err_msg,
                "*** Can't resolve parent name for `%.24s' unambiguously! ***",
		       add_record->ufn);

      add_error_to_request_rec(request_record,
                               add_record->baseobject,
                               QERR_cannot_resolve_name,
                               (struct DSError *) NULL,
                               err_msg);

      add_record->result = add_result_alloc();
      
      add_record->result->errors = request_record->errors;
      request_record->errors = NULLError;

      add_record->result->was_successful = FALSE;
      
      add_record->baseobject = copy_string(baseobject);
      
      ufname_result_free(&results);

      return RQ_error_returned;
    }
} /* process_add_resolve_request_result */


/*
 *	Free'ing procedures
 *
 *
 */
void free_add_result(result_ptr)
     addResult *result_ptr;
{
  addResult result = *result_ptr;

  if (result == NULLAddResult)
    return;

  if (result->baseobject != NULLCP)
    (void) free(result->baseobject);

  if (result->errors != NULLError)
    error_list_free(&result->errors);

  (void) free((char *) result);

  *result_ptr = NULLAddResult;
} /* free_add_result */


void free_add_rec(record)
     addRec record;
{
  if (record == NULLAddRec)
    return;

  if (record->baseobject != NULLCP)
    (void) free(record->baseobject);

  if (record->ufn != NULLCP)
    (void) free(record->ufn);

  if (record->result != NULLAddResult)
    free_add_result(&record->result);

  (void) free((char *) record);
} /* free_add_rec */

