#ifndef lint
static char *rcsid = "$Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/delete.c,v 0.95 91/09/02 11:33:09 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/delete.c,v 0.95 91/09/02 11:33:09 emsrdsm Exp Locker: emsrdsm $ */

/*****************************************************************************

  delete.c - Delete directory entries

*****************************************************************************/

#include "types.h"
#include "util.h"
#include "read.h"
#include "add.h"
#include "request.h"
#include "ds_tasks.h"

#include "quipu/common.h"
#include "quipu/entry.h"
#include "quipu/add.h"
#include "quipu/name.h"
#include "quipu/util.h"
#include "quipu/dap2.h"
#include "quipu/attrvalue.h"


/*
 *	Static procs...
 */

static errorList start_delete();

/*
 *	Externs..
 */

#ifndef NO_STATS
extern LLog *log_stat;
#endif


/*
 * - get_delete_result() -
 *
 *
 */
deleteResult get_delete_result(request_id)
     QCardinal request_id;
{
  requestRec request = _get_request_of_id(request_id);
  deleteRec delrec;
  deleteResult result;

  if (request == NULLReqRec)
    return NULLDeleteResult;

  delrec = request->DELETE_REC;
  result = delrec->result;

  delrec->result = NULLDeleteResult;

  _request_complete(request_id);

  return result;
} /* get_delete_result */


/*
 * - do_delete() -
 *
 *
 */
errorList do_delete(id_ptr, baseobject)
     QCardinal *id_ptr;
     char *baseobject;
{
  deleteRec delete_request;
  requestRec this_request;

  (void) _request_invoked(DELETE_ENTRY, id_ptr);
  this_request = _get_request_of_id(*id_ptr);

  delete_request = this_request->DELETE_REC;

  delete_request->request_id = *id_ptr;
  delete_request->delete_object = copy_string(baseobject);
  
  return start_delete(*id_ptr, baseobject, &delete_request->task_id);
} /* do_add */

/*
 * - start_delete() -
 *
 *
 */
static errorList start_delete(request_id, baseobject, task_id_ptr)
     QCardinal request_id;
     char *baseobject;
     int *task_id_ptr;
{
  int task_id;

  struct ds_removeentry_arg rem_arg;
  struct DAPindication di;
  
  errorList errors;
  char err_msg[LINESIZE];

  (void) get_default_service(&rem_arg.rma_common);

  if ((rem_arg.rma_object = str2dn(baseobject)) == NULLDN)
    {
      (void) sprintf(err_msg, "*** Bad entry name given `%s'", baseobject);
      add_error_to_list(&errors,
                        NULLCP,
                        QERR_bad_name,
                        (struct DSError *) NULL,
                        err_msg);

      _request_complete(request_id);

      return errors;
    }

  if (_task_invoked(REMOVE_TASK, baseobject, request_id, &task_id) != QERR_ok)
    {
      add_error_to_list(&errors,
                        NULLCP,
                        QERR_internal_limit_reached,
                        (struct DSError *) NULL,
                        NULLCP);

      dn_free(rem_arg.rma_object);

      _request_complete(request_id);

      return errors;
    }

  if (DapRemoveEntry(dsap_ad, task_id, &rem_arg, &di, ROS_ASYNC) == NOTOK)
    {
      add_error_to_list(&errors,
                        NULLCP,
                        QERR_internal,
                        (struct DSError *) NULL,
                        NULLCP);

      dn_free(rem_arg.rma_object);

      _request_complete(request_id);

      return errors;
    }

#ifndef NO_STATS
  LLOG(log_stat,LLOG_NOTICE,("DELETE ENTRY +%s, task %d", baseobject,task_id));
#endif

  dn_free(rem_arg.rma_object);

  *task_id_ptr = task_id;

  return NULLError;
} /* start_delete */


/*
 * - process_delete_ds_result() -
 *
 *
 */
request_state process_delete_ds_result(request, task_id)
     requestRec request;
     int task_id;
{
  deleteRec delrec = request->DELETE_REC;

    /* This shouldn't happen */
  if (_get_task_of_id(task_id) == NULLDsTask)
    return RQ_processing;

  delrec->result = delete_result_alloc();
  
  delrec->result->was_successful = TRUE;
  delrec->result->errors = NULLError;
  
  return RQ_results_returned;
} /* process_delete_ds_result */


/*
 * - process_delete_ds_error() -
 *
 *
 */
request_state process_delete_ds_error(request, task_id, ds_error)
     requestRec request;
     int task_id;
     struct DSError *ds_error;
{
  DsTask task_rec;
  deleteRec delrec = request->DELETE_REC;
  QE_error_code error_type;

  /* This shouldn't happen */
  if ((task_rec = _get_task_of_id(task_id)) == NULLDsTask)
    return RQ_processing;

  /* Save error and log it */
  error_type = get_log_error_type(ds_error, task_id);
  add_error_to_request_rec(request,
                           task_rec->baseobject,
                           error_type,
                           ds_error,
                           NULLCP);

  delrec->result = delete_result_alloc();
  delrec->result->was_successful = FALSE;

  delrec->result->errors = request->errors;
  request->errors = NULLError;
  
  return RQ_error_returned;
} /* process_add_ds_error */

/*
 *  Freeing procs...
 */

void free_delete_rec(record)
     deleteRec record;
{
  if (record == NULLDeleteRec)
    return;

  if (record->delete_object != NULLCP)
    (void) free(record->delete_object);
  
  if (record->result != NULLDeleteResult)
    free_delete_result(&record->result);

  (void) free((char *) record);
} /* free_delete_rec */


void free_delete_result(result_ptr)
     deleteResult *result_ptr;
{
  deleteResult result= *result_ptr;

  if (result == NULLDeleteResult)
    return;

  if (result->errors != NULLError)
    error_list_free(&result->errors);

  (void) free((char *) result);
  
  *result_ptr = NULLDeleteResult;
} /* free_delete_result */
