#ifndef lint
static char *rcsid = "$Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/rename.c,v 0.95 91/09/02 11:34:44 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/rename.c,v 0.95 91/09/02 11:34:44 emsrdsm Exp Locker: emsrdsm $ */

/*****************************************************************************

  rename.c - Rename directory entries

*****************************************************************************/

#include "types.h"
#include "util.h"
#include "ufname.h"
#include "rename.h"
#include "request.h"
#include "ds_tasks.h"

#include "quipu/common.h"
#include "quipu/entry.h"
#include "quipu/name.h"
#include "quipu/modifyrdn.h"
#include "quipu/util.h"
#include "quipu/dap2.h"
#include "quipu/attrvalue.h"
#include "quipu/photo.h"

/*
 *	Static procs..
 */

static errorList start_rename();

static request_state process_resolve_result(),
  		     process_add_entry_result(),
		     process_delete_entry_result(),
                     process_read_result();

static void ufname_cmp();

/*
 *	Externs..
 */

#ifndef NO_STATS
extern LLog *log_stat;
#endif


/*
 * - get_rename_result() -
 *
 *
 */
renameResult get_rename_result(request_id)
     QCardinal request_id;
{
  requestRec request_record;
  renameResult result;

  if ((request_record = _get_request_of_id(request_id)) == NULLReqRec)
    return NULLRenameResult;

  result = request_record->RENAME_REC->result;

  request_record->RENAME_REC->result = NULLRenameResult;

  _request_complete(request_id);

  return result;
} /* get_rename_result */

/*
 * - do_rename() -
 *
 *
 */
errorList do_rename(id_ptr, baseobject, new_name)
     QCardinal *id_ptr;
     char *baseobject, *new_name;
{
  renameRec rename_request;
  requestRec this_request;


  (void) _request_invoked(RENAME_ENTRY, id_ptr);
  this_request = _get_request_of_id(*id_ptr);

  rename_request = this_request->RENAME_REC;

  rename_request->request_id = *id_ptr;
  rename_request->task_id = NO_TASK;

  rename_request->baseobject = copy_string(baseobject);
  rename_request->new_name = copy_string(new_name);

  rename_request->entry_attrs = NULLModifyAttr;

  rename_request->status = RN_init;
  
  return start_rename(*id_ptr, baseobject, new_name, &rename_request->task_id);
} /* do_rename */

/*
 * - start_rename() -
 *
 *
 */
static errorList start_rename(request_id, baseobject, new_name, task_id_ptr)
     QCardinal request_id;
     char *baseobject, *new_name;
     int *task_id_ptr;
{
  QE_error_code resolve_status;

  errorList errors = NULLError;
  requestRec request_record = _get_request_of_id(request_id);
  renameRec rename_record = request_record->RENAME_REC;

  char *new_name_parent;
  QBool need_reparent, new_rdn = FALSE;

  char err_msg[LINESIZE], old_name[LINESIZE];

  namePart new_ufname;
  
  extern int dsap_ad;

  friendlify(baseobject, old_name);

  if ((new_ufname = str2ufname(new_name)) == NULLNamePart)
    {
      (void) sprintf(err_msg,
		     "*** Bad syntax in entry name `%s' ***",
		     new_name);

      add_error_to_list(&errors, 
			NULLCP,
			QERR_bad_name,
			(struct DSError *) NULL,
			err_msg);

      _request_complete(request_id);

      return errors;
    }

  name_part_free(&new_ufname);

  ufname_cmp(old_name, new_name, &new_rdn, &need_reparent);


  if (new_rdn == FALSE && need_reparent == FALSE)
    {
      (void) sprintf(err_msg,
		     "*** No changes to entry name `%s' required! ***",
		     new_name);

      add_error_to_list(&errors,
			NULLCP,
			QERR_no_name_changes,
			(struct DSError *) NULL,
			err_msg);

      _request_complete(request_id);

      return errors;
    }


  if (need_reparent == TRUE)
    {
      /*
       *
       *	First have to resolve the parent of the new entry
       *
       */
      
      for (new_name_parent = new_name;
	   !isnull(*new_name_parent);
	   new_name_parent++)
	if (*new_name_parent == '\"')
	  while (*new_name_parent != '\"' && !isnull(*new_name_parent))
	    new_name_parent++;
	else if (*new_name_parent == ',')
	  {
	    new_name_parent++;
	    
	    while (isspace(*new_name_parent))
	      new_name_parent++;
	    break;
	  }

      
      if ((resolve_status = do_ufn_resolve(NULLEntryList,
					   new_name_parent,
					   might_not,
					   &(rename_record->sub_request_id)))
	  != QERR_ok)
	{
	  _request_complete(request_id);

	  switch (resolve_status)
	    {
	    case QERR_bad_value_syntax:
	      (void) sprintf(err_msg,
			     "*** Bad syntax in name `%.32s...' ! ***",
			     new_name);
	      
	      add_error_to_list(&errors,
				NULLCP,
				QERR_bad_value_syntax,
				(struct DSError *) NULL,
				err_msg);
	      
	      return errors;
	      
	    case QERR_internal_limit_reached:
	      add_error_to_list(&errors,
				NULLCP,
				QERR_bad_value_syntax,
				(struct DSError *) NULL,
			"*** Internal limit reached, can't perform! ***");
	      
	      return errors;

            default:
	      add_error_to_list(&errors,
				NULLCP,
				QERR_internal,
				(struct DSError *) NULL,
				"*** Internal error, can't perform! ***");

	      return errors;
	    }
	}

      tag_as_sub_request(rename_record->sub_request_id, request_id);
  
      rename_record->status = RN_resolve;
    }
  else
    {
      int task_id;
      
      struct ds_modifyrdn_arg mod_arg;
      struct DAPindication di;

      char str_rdn[LINESIZE], *str;
      
      (void) get_default_service(&mod_arg.mra_common);

      mod_arg.mra_common.ca_servicecontrol.svc_options =
	mod_arg.mra_common.ca_servicecontrol.svc_options | SVC_OPT_PREFERCHAIN;

      mod_arg.mra_object = str2dn(baseobject);
      
      /*
       *	Compose the new rdn from baseobject and new_name
       */

      for (str = baseobject; !isnull(*str); str++)
	;
      
      while (*str != '@' && str > baseobject)
	str--;
      
      if (*str == '@')
	str++;

      (void) strcpy(str_rdn, str);

      str = index(str_rdn, '=');

      *++str = '\0';
      
      (void) strcat(str_rdn, new_name);

      str = str_rdn;
      while (*str != ',' && !isnull(*str))
	if (*str == '\"')
	  while (*str++ != '\"' && !isnull(*str))
	    ;
	else
	  str++;
      
      *str = '\0';
      
      mod_arg.mra_newrdn = str2rdn(str_rdn);
      mod_arg.deleterdn = TRUE;


      /*
       *	Record task invocation
       */

      if (_task_invoked(MODIFY_RDN_TASK, baseobject, request_id, &task_id)
	  != QERR_ok)
	{
	  add_error_to_list(&errors,
			    NULLCP,
			    QERR_internal_limit_reached,
			    (struct DSError *) NULL,
			    NULLCP);

	  _request_complete(request_id);
	  
	  rdn_free(mod_arg.mra_newrdn);
	  dn_free(mod_arg.mra_object);

	  return errors;
	}

      *task_id_ptr = task_id;
      
      if (DapModifyRDN(dsap_ad, task_id, &mod_arg, &di, ROS_ASYNC) == NOTOK)
	{
	  add_error_to_list(&errors,
			    NULLCP,
			    QERR_request_failed,
			    (struct DSError *) NULL,
			    NULLCP);
	  
	  _request_complete(request_id);

	  _task_complete(task_id);

	  rdn_free(mod_arg.mra_newrdn);
	  dn_free(mod_arg.mra_object);
	  
	  return errors;
	}

#ifndef NO_STATS
      LLOG (log_stat, LLOG_NOTICE,
	    ("MODIFY RDN of +%s to %s, task %d",
	     baseobject,
	     str_rdn,
	     task_id));
#endif

    }

  return NULLError;
} /* start_rename */

/*
 * - process_rename_ds_result() -
 *
 *
 */
request_state process_rename_ds_result(request, task_id)
     requestRec request;
     int task_id;
{
  renameRec renrec = request->RENAME_REC;

  /* This shouldn't happen */
  if (_get_task_of_id(task_id) == NULLDsTask)
    return RQ_processing;

  renrec->result = rename_result_alloc();

  renrec->result->was_successful = TRUE;

  renrec->result->errors = request->errors;
  request->errors = NULLError;
  
  renrec->result->updated_object = copy_string(renrec->new_baseobject);
  renrec->result->status = RenameSuccessful;

  /* Remove invocation record for this task */
  _task_complete(task_id);

  return RQ_results_returned;
} /* process_rename_ds_result */

/*
 * - process_rename_ds_error() -
 *
 *
 */
request_state process_rename_ds_error(request, task_id, ds_error)
     requestRec request;
     int task_id;
     struct DSError *ds_error;
{
  DsTask task_rec;
  renameRec renrec = request->RENAME_REC;
  QE_error_code error_type;

  /* This shouldn't happen */
  if ((task_rec = _get_task_of_id(task_id)) == NULLDsTask)
    return RQ_processing;

  /* Save error and log it */
  error_type = get_log_error_type(ds_error, task_id);
  add_error_to_request_rec(request,
                           task_rec->baseobject,
                           error_type,
                           ds_error,
                           NULLCP);

  renrec->result = rename_result_alloc();

  renrec->result->was_successful = FALSE;

  renrec->result->errors = request->errors;
  request->errors = NULLError;
  
  renrec->result->updated_object = NULLCP;
  renrec->result->status = ModifyRDNFailed;


  /* Remove invocation record for this task */
  _task_complete(task_id);

  return RQ_error_returned;
} /* process_rename_ds_error */

/*
 * - process_rename_request_result() -
 *
 *
 */
request_state process_rename_request_result(sub_request_id)
     QCardinal sub_request_id;
{
  QCardinal request_id;
  requestRec request_record, sub_request_record;

  sub_request_record = _get_request_of_id(sub_request_id);
  request_id = sub_request_record->owner_request_id;

  request_record = _get_request_of_id(request_id);


  switch (sub_request_record->type)
    {
    case UFNAME:
      
      return process_resolve_result(request_record, sub_request_id);

    case ADD_ENTRY:

      return process_add_entry_result(request_record, sub_request_id);

    case DELETE_ENTRY:
      
      return process_delete_entry_result(request_record, sub_request_id);

    case ENTRY_TEMPLATE:
      
      return process_read_result(request_record, sub_request_id);
    }

  return RQ_processing;
} /* process_rename_request_result */


/*
 * - process_add_entry_result() -
 *
 *
 */
static request_state process_add_entry_result(request_record, sub_request_id)
     requestRec request_record;
     QCardinal sub_request_id;
{
  addResult result;
  renameRec rename_record = request_record->RENAME_REC;
  
  errorList errors, save_errors;
  char err_msg[LINESIZE], old_name[LINESIZE];

  QCardinal new_sub_request_id;

  result = get_add_result(sub_request_id);

  if (result->was_successful == FALSE)
    {
      for (errors = result->errors;
	   errors != NULLError && errors->next != NULLError;
	   errors = errors->next)
	if (errors->ds_message != NULLCP)
	  {
	    (void) sprintf(err_msg,
			   "While adding entry `%.32s...'\n",
			   rename_record->new_name);
	    
	    (void) strcat(err_msg, errors->ds_message);
	    
	    (void) free(errors->ds_message);
	    errors->ds_message = copy_string(err_msg);
	  }

      if (errors != NULLError)
	{
	  if (errors->ds_message != NULLCP)
	    {
	      (void) sprintf(err_msg,
			     "While adding entry `%.32s...'\n",
			     rename_record->new_name);
	      
	      (void) strcat(err_msg, errors->ds_message);
	      
	      (void) free(errors->ds_message);
	      errors->ds_message = copy_string(err_msg);
	    }	  

	  errors->next = request_record->errors;
	  request_record->errors = result->errors;
	}

      result->errors = NULLError;
      free_add_result(&result);

      rename_record->result = rename_result_alloc();

      rename_record->result->was_successful = FALSE;
      rename_record->result->errors = request_record->errors;

      request_record->errors = NULLError;
      rename_record->result->status = AddFailed;

      rename_record->result->updated_object = NULLCP;
      
      return RQ_error_returned;
    }
  else
    {
      free_add_result(&result);
      
      rename_record->status = RN_delete;

      if ((errors = do_delete(&new_sub_request_id, rename_record->baseobject))
	  != NULLError)
	{
	  save_errors = errors;

	  friendlify(rename_record->baseobject, old_name);

	  while (errors != NULLError)
	    {
	      if (errors->ds_message != NULLCP)
		{
		  (void) sprintf(err_msg,
                         "While deleting entry `%.32s...'\n",
                         old_name);

		  (void) strcat(err_msg, errors->ds_message);
		  
		  (void) free(errors->ds_message);
		  errors->ds_message = copy_string(err_msg);
		}

	      if (errors->next == NULLError)
		{
		  errors->next = request_record->errors;
		  break;
		}
	    }

	  if (save_errors != NULLError)
	    request_record->errors = save_errors;

	  rename_record->status = RN_complete;
	  
	  rename_record->result = rename_result_alloc();
	  
	  rename_record->result->was_successful = FALSE;
	  rename_record->result->errors = request_record->errors;
	  
	  request_record->errors = NULLError;
	  rename_record->result->status = DeleteFailed;
	  
	  rename_record->result->updated_object = NULLCP;

	  return RQ_error_returned;
	}
      
      /*
       *   Delete operation submitted
       */

      tag_as_sub_request(new_sub_request_id, rename_record->request_id);
      rename_record->sub_request_id = new_sub_request_id;

      return RQ_processing;
    }
} /* process_add_entry_result */


/*
 * - process_delete_entry_result() -
 *
 *
 */
static request_state process_delete_entry_result(request_record,sub_request_id)
     requestRec request_record;
     QCardinal sub_request_id;
{
  deleteResult result;
  renameRec rename_record = request_record->RENAME_REC;

  errorList errors = NULLError;
  char err_msg[LINESIZE], old_name[LINESIZE];

  result = get_delete_result(sub_request_id);

  friendlify(rename_record->baseobject, old_name);

  if (result->was_successful == FALSE)
    {
      for (errors = result->errors;
           errors != NULLError && errors->next != NULLError;
           errors = errors->next)
        if (errors->ds_message != NULLCP)
          {
            (void) sprintf(err_msg,
                           "While deleting entry `%.32s...'\n",
			   old_name);

	    (void) strcat(err_msg, errors->ds_message);

            (void) free(errors->ds_message);
            errors->ds_message = copy_string(err_msg);
          }

      if (errors != NULLError)
        {
          if (errors->ds_message != NULLCP)
            {
              (void) sprintf(err_msg,
                             "While deleting entry `%.32s...'\n",
                             old_name);
	      
	      (void) strcat(err_msg, errors->ds_message);

              (void) free(errors->ds_message);
              errors->ds_message = copy_string(err_msg);
            }

          errors->next = request_record->errors;
          request_record->errors = result->errors;
	  
	  result->errors = NULLError;
        }

      free_delete_result(&result);

      rename_record->status = RN_complete;

      rename_record->result = rename_result_alloc();

      rename_record->result->was_successful = FALSE;
      rename_record->result->errors = request_record->errors;

      request_record->errors = NULLError;
      rename_record->result->status = DeleteFailed;

      rename_record->result->updated_object = NULLCP;

      return RQ_error_returned;
    }
  else
    {
      free_delete_result(&result);

      rename_record->status = RN_complete;

      rename_record->result = rename_result_alloc();

      rename_record->result->was_successful = TRUE;
      rename_record->result->errors = request_record->errors;

      request_record->errors = NULLError;
      rename_record->result->status = RenameSuccessful;

      rename_record->result->updated_object =
	copy_string(rename_record->new_baseobject);
      

      return RQ_results_returned;
    }
} /* process_delete_entry_result */

/*
 * - process_resolve_result() -
 *
 *
 */
static request_state process_resolve_result(request_record, sub_request_id)
     requestRec request_record;
     QCardinal sub_request_id;
{
  ufnResults results;
  renameRec rename_record = request_record->RENAME_REC;

  errorList errors = NULLError;
  char err_msg[LINESIZE],
       new_baseobject[LINESIZE],
       buf[LINESIZE],
       old_name[LINESIZE],
       new_name[LINESIZE];

  QCardinal new_sub_request_id;
  char *str;

  QE_error_code sub_request_status;
  modifyAttr request_attrs = NULLModifyAttr;
  
  QBool new_rdn, need_reparent;


  results = get_ufn_results(sub_request_id);

  /*
   * Will have to change this if poor matches are to be queried
   */

  _request_complete(sub_request_id);


  /*
   *	Put errors into parent request record
   */

  for (errors = results->errors; errors != NULLError; errors = errors->next)
      if (errors->ds_message != NULLCP)
	{
	  (void) sprintf(err_msg,
			 "While resolving name `%.32s...'\n",
			 rename_record->new_name);
	  
	  (void) strcat(err_msg, errors->ds_message);

	  (void) free(errors->ds_message);
	  errors->ds_message = copy_string(err_msg);
	}

  request_record->errors = results->errors;
  results->errors = NULLError;

  /*
   *	Has name resolution succeded (matches = 1) ?
   */

  if (results->match_status == GoodMatches && results->match_num == 1)
    {
      (void) strcpy(new_baseobject, results->matches->string_dn);

      friendlify(new_baseobject, new_name);
      friendlify(rename_record->baseobject, old_name);

      for (str = old_name; !isnull(*str) && *str != ','; str++)
	if (*str == '\"')
	  {
	    while (!isnull(*++str) && *str != '\"')
	      ;
	    
	    if (isnull(*str))
	      break;
	  }

      if (isnull(*str))
	need_reparent = FALSE;
      else
	{
	  while (!isnull(*str) && !isalpha(*str))
	    str++;

	  ufname_cmp(new_name, str, &new_rdn, &need_reparent);
	}

      if (need_reparent == FALSE && new_rdn == FALSE)
	{
	  ufname_result_free(&results);

	  add_error_to_request_rec(request_record,
				   rename_record->baseobject,
				   QERR_no_name_changes,
				   (struct DSError *) NULL,
				   NULLCP);

	  rename_record->status = RN_complete;
	  
	  rename_record->result = rename_result_alloc();
	  
	  rename_record->result->was_successful = FALSE;
	  rename_record->result->status = AddFailed;
	  
	  rename_record->result->errors = request_record->errors;
	  request_record->errors = NULLError;

	  rename_record->result->updated_object = NULLCP;

	  return RQ_error_returned;
	}

      for (str = rename_record->baseobject; !isnull(*str); str++)
	;
      
      while (*str != '@' && str > rename_record->baseobject)
	str--;

      (void) strcpy(buf, str);

      str = index(buf, '=');
      *++str = '\0';

      (void) strcat(buf, rename_record->new_name);
      
      for (str = buf; !isnull(*str); str++)
	if (*str == '\"')
	  while (*str != '\"' && !isnull(*str))
	    str++;
	else if (*str == ',')
	  {
	    str--;

	    while (isspace(*str))
	      str--;

	    *++str = '\0';
	    
	    break;
	  }

      
      (void) strcat(new_baseobject, buf);

      rename_record->new_baseobject = copy_string(new_baseobject);

      ufname_result_free(&results);

      rename_record->status = RN_read;

      if ((sub_request_status = do_make_template(rename_record->baseobject,
						 &new_sub_request_id,
						 &request_attrs))
	   != QERR_ok)
	{
	  add_error_to_request_rec(request_record,
				   rename_record->baseobject,
				   sub_request_status,
				   (struct DSError *) NULL,
				   NULLCP);

	  rename_record->status = RN_complete;
	  
	  rename_record->result = rename_result_alloc();
	  
	  rename_record->result->was_successful = FALSE;
	  rename_record->result->status = ReadFailed;
	  
	  rename_record->result->errors = request_record->errors;
	  request_record->errors = NULLError;

	  rename_record->result->updated_object = NULLCP;

	  return RQ_error_returned;
	}

      /*
       *     Make template request initiated
       */

      rename_record->sub_request_id = new_sub_request_id;
      tag_as_sub_request(new_sub_request_id, rename_record->request_id);
      
      return RQ_processing;
    }
  else
    /*
     *	Name resolution failed
     */
    {
      ufname_result_free(&results);

      (void) sprintf(err_msg,
		"*** Can't resolve new parent name `%.24s' unambiguously! ***",
		     rename_record->new_name);

      add_error_to_request_rec(request_record,
			       rename_record->baseobject,
			       QERR_unable_to_proceed,
			       (struct DSError *) NULL,
			       err_msg);
      
      rename_record->result = rename_result_alloc();

      rename_record->result->was_successful = FALSE;
      rename_record->result->status = ResolveFailed;

      rename_record->result->errors = request_record->errors;
      request_record->errors = NULLError;

      rename_record->result->updated_object = NULLCP;

      return RQ_error_returned;
    }
} /* process_resolve_result */


/*
 * - process_read_result() -
 *
 *
 */
static request_state process_read_result(request_record, sub_request_id)
     requestRec request_record;
     QCardinal sub_request_id;
{
  makeTemplateResult result;
  renameRec rename_record = request_record->RENAME_REC;
  
  errorList errors, save_errors;
  char err_msg[LINESIZE];

  QCardinal new_sub_request_id;

  modifyAttr curr_attr;

  result = get_make_template_result(sub_request_id);

  if (result->errors != NULLError && result->must_exc == NULLModifyAttr)
    /*
     *	Make template failed
     */
    {
      save_errors = errors = result->errors;
      result->errors = NULLError;

      while (errors != NULLError)
	{
	  if (errors->ds_message != NULLCP)
	    {
	      (void) sprintf(err_msg,
			     "While reading entry `%.32s...'\n",
			     rename_record->baseobject);
	      
	      (void) strcat(err_msg, errors->ds_message);
	      
	      (void) free(errors->ds_message);
	      errors->ds_message = copy_string(err_msg);
	    }

	  if (errors->next = NULLError)
	    {
	      errors->next = request_record->errors;
	      request_record->errors = save_errors;
	      
	      break;
	    }
	}
      
      free_make_template_result(&result);
      
      rename_record->status = RN_complete;

      rename_record->result = rename_result_alloc();

      rename_record->result->was_successful = FALSE;
      rename_record->result->errors = request_record->errors;

      request_record->errors = NULLError;
      rename_record->result->status = ReadFailed;

      rename_record->result->updated_object = NULLCP;

      return RQ_error_returned;
    }
  else
    /*
     *	Succeeded, now try to add entry against new name
     */
    {
      rename_record->status = RN_add;

      rename_record->entry_attrs = result->must_exc;

      for (curr_attr = result->must_exc;
	   curr_attr != NULLModifyAttr && curr_attr->next != NULLModifyAttr;
	   curr_attr = curr_attr->next)
	;

      if (curr_attr != NULLModifyAttr)
	curr_attr->next = result->may_exc;

      rename_record->entry_attrs = result->must_exc;
      result->must_exc = result->may_exc = NULLModifyAttr ;

      free_make_template_result(&result);

      if ((errors = do_add(&new_sub_request_id,
			   rename_record->new_baseobject,
			   NULLCP,
			   NULLAttrT,
			   rename_record->entry_attrs)) != NULLError)
	{
	  save_errors = errors;

	  while (errors != NULLError)
	    {
	      if (errors->ds_message != NULLCP)
		{
		  (void) sprintf(err_msg,
				 "While adding entry `%.32s...'\n",
				 rename_record->new_name);
		  
		  (void) strcat(err_msg, errors->ds_message);
		  
		  (void) free(errors->ds_message);
		  errors->ds_message = copy_string(errors->ds_message);
		}
	      
	      if (errors->next == NULLError)
		{
		  errors->next = request_record->errors;
		  break;
		}
	    }

	  if (save_errors != NULLError)
	    request_record->errors = save_errors;

	  rename_record->status = RN_complete;
	  
	  rename_record->result = rename_result_alloc();
	  
	  rename_record->result->was_successful = FALSE;
	  rename_record->result->errors = request_record->errors;
	  
	  request_record->errors = NULLError;
	  rename_record->result->status = AddFailed;

	  rename_record->result->updated_object = NULLCP;

	  return RQ_error_returned;
	}

      rename_record->sub_request_id = new_sub_request_id;
      tag_as_sub_request(new_sub_request_id, rename_record->request_id);
      
      return RQ_processing;
   }
} /* process_read_result */

/*
 * 	Freeing procs...
 *
 *
 */

void free_rename_rec(record)
     renameRec record;
{
  if (record == NULLRenameRec)
    return;

  if (record->baseobject != NULLCP)
    (void) free(record->baseobject);

  if (record->new_name != NULLCP)
    (void) free(record->new_name);

  if (record->new_baseobject != NULLCP)
    (void) free(record->new_baseobject);

  if (record->result != NULLRenameResult)
    free_rename_result(&(record->result));

  (void) free((char *) record);
} /* free_rename_rec */


void free_rename_result(result_ptr)
     renameResult *result_ptr;
{
  renameResult result = *result_ptr;

  if (result == NULLRenameResult)
    return;

  if (result->errors != NULLError)
    error_list_free(&(result->errors));

  if (result->updated_object != NULLCP)
    (void) free(result->updated_object);

  (void) free((char *) result);

  *result_ptr = NULLRenameResult;
} /* free_rename_result */


/*
 * Misc utilities
 *
 *
 */

static void ufname_cmp(name1_str, name2_str, rdn_diff, parent_diff)
     char *name1_str, *name2_str;
     QBool *rdn_diff, *parent_diff;
{
  namePart name1, name2, save1, save2; 
  QBool rdn = FALSE, parent = FALSE;

  save1 = str2ufname(name1_str);
  save2 = str2ufname(name2_str);

  for (name2 = save2, name1 = save1;
       name2 != NULLNamePart && name1 != NULLNamePart;
       name2 = name2->next, name1 = name1->next)
    if (lexequ(name2->part_name, name1->part_name) != 0)
      {
	if (name2->next == NULLNamePart && name1->next == NULLNamePart)
	  rdn = TRUE;
	else
	  parent = TRUE;
      }

  if (name2 != NULLNamePart || name1 != NULLNamePart)
    {
      parent = TRUE;
      rdn = TRUE;
    }

  name_part_free(&save1);
  name_part_free(&save2);

  *parent_diff = parent;
  *rdn_diff = rdn;
} /* ufname_cmp */
