#ifndef lint
static char *rcsid = "$Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/ufsearch.c,v 0.95 91/09/02 11:35:08 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/query_engine/RCS/ufsearch.c,v 0.95 91/09/02 11:35:08 emsrdsm Exp Locker: emsrdsm $ */

/*****************************************************************************

  ufsearch.c - Implements unsequenced attribute based entry search

*****************************************************************************/


#include "types.h"
#include "util.h"
#include "ufsearch.h"
#include "ds_tasks.h"
#include "request.h"
#include "ds_util.h"
#include "read.h"
#include "ds_tasks.h"


#include "quipu/ds_search.h"
#include "quipu/dap2.h"
#include "quipu/name.h"


/*
 *	Externals...
 */

extern int dsap_ad;
extern int next_task_id;

/* extern ufsPath ufs_paths; */

extern char *TidyString();

#ifndef NO_STATS
extern LLog    *log_stat;
#endif


/*
 *	Local procs....
 */

static errorList process_ufsearch();
static errorList do_directory_search();
static errorList follow_up();
static errorList make_search_filters();
static request_state check_ufsearch_state();

static AttributeType check_get_entry_oc();

static QBool get_default_path();
static QBool is_target_class();
static stringCell make_sequence_from_values();


/*
 *	Local data
 */

static searchClass search_classes[MAX_Search_Classes];
static QCardinal search_class_num = 0;


/*
 *	Global data
 */

targetClass target_classes[MAX_Target_Classes];
QCardinal target_class_num = 0;


/*
 * - get_ufsearch_result() -
 *
 *
 */
ufsearchResult get_ufsearch_result(request_id)
     QCardinal request_id;
{
  requestRec request_record = _get_request_of_id(request_id);
  ufsearchRec ufsrec = request_record->UFSEARCH_REC;
  ufsearchResult result;

  result = ufsrec->result;

  ufsrec->result = NULLUfsearchResult;

  _request_complete(request_id);

  return result;
} /* get_ufsearch_result */

/*
 * - get_ufsearch_report() -
 *
 *
 */
ufsearchReport get_ufsearch_report(request_id)
     QCardinal request_id;
{
  requestRec request_record = _get_request_of_id(request_id);
  ufsearchRec ufsrec = request_record->UFSEARCH_REC;
  ufsearchReport report;

  report = ufsrec->report;

  ufsrec->report = NULLUfsearchReport;
  
  return report;
} /* get_ufsearch_report */


/*
 * - do_ufsearch() -
 *
 *
 */
errorList do_ufsearch(target_data, context_data, id_ptr)
     searchData target_data, context_data;
     QCardinal *id_ptr;
{
  requestRec request;
  ufsearchRec ufsrec;
  searchData curr_srch_data;
  entryList path = NULLEntryList;
  errorList errors = NULLError;

  /*
   *	Check that search data is valid
   */

  if (get_default_path(target_data, context_data, &path) == FALSE)
    {
      add_error_to_list(&errors,
			NULLCP,
			QERR_invalid_search_class,
			(struct DSError *) NULL,
			NULLCP);
      
      return errors;
    }

  
  /*
   * Record request invocation
   */

  (void) _request_invoked(UFSEARCH, id_ptr);

  request = _get_request_of_id(*id_ptr);
  ufsrec = request->UFSEARCH_REC;

  ufsrec->trying_target_search = FALSE;
  ufsrec->made_good_matches = FALSE;

  ufsrec->target_data = target_data;
  ufsrec->context_data = context_data;


  for (curr_srch_data = context_data;
       curr_srch_data != NULLSearchData;
       curr_srch_data = curr_srch_data->next)
    {
      curr_srch_data->attempted = curr_srch_data->used = FALSE;
      curr_srch_data->task_list = NULLDsTask;
    }

  target_data->attempted = target_data->used = FALSE;
  target_data->task_list = NULLDsTask;

  
  if (path != NULLEntryList)
    ufsrec->default_path = path->next;

  ufsrec->to_follow = path;
  path->next = NULLEntryList;
  
  if ((errors = process_ufsearch(ufsrec)) != NULLError)
    {
      _request_complete(*id_ptr);
      return errors;
    }

  return NULLError;
} /* do_ufsearch */
     

/*
 * - process_ufsearch() -
 *
 *
 */
static errorList process_ufsearch(ufsrec)
     ufsearchRec ufsrec;
{
  entryList to_follow = ufsrec->to_follow, default_path = ufsrec->default_path;
  searchData curr_srch_data;
  errorList error = NULLError;
  

  /*
   *	If all search data has been used then search has failed
   */

  for (curr_srch_data = ufsrec->context_data;
       curr_srch_data != NULLSearchData && curr_srch_data->used == TRUE;
       curr_srch_data = curr_srch_data->next)
    ;

  if (curr_srch_data == NULLSearchData)
    {
      /*
       *	All context data has been attempted....
       */
      if (ufsrec->target_data->attempted == TRUE)
	/*
	 *  .... and target search has been done!
	 */
	{
	  /*
	   *	Try next element in search path
	   */

	  if (default_path != NULLEntryList &&
	      ufsrec->made_good_matches == FALSE)
	    {
	      if (to_follow != NULLEntryList)
		dn_list_free(&to_follow);

	      to_follow = ufsrec->to_follow = default_path;
	      default_path = ufsrec->default_path = default_path->next;

	      to_follow->next = NULLEntryList;
	      

	      for (curr_srch_data = ufsrec->context_data;
		   curr_srch_data != NULLSearchData;
		   curr_srch_data = curr_srch_data->next)
		{
		  curr_srch_data->attempted = curr_srch_data->used = FALSE;
		  ds_task_list_free(&curr_srch_data->task_list);
		}
	      
	      ufsrec->target_data->attempted =
		ufsrec->target_data->used =
		  FALSE;

	      ds_task_list_free(&ufsrec->target_data->task_list);


	      if ((error = follow_up(ufsrec->target_data, to_follow, ufsrec))
		  != NULLError)
		{
		  if (error->error_type == QERR_invalid_search_class)
		    {
		      error->error_type = QERR_nothing_found;
		      return error;
		    }
		  else
		    return error;
		}
	      
	      dn_list_free(&to_follow);
	      ufsrec->to_follow = NULLEntryList;

	      return NULLError;
	    }


	  /*
	   *	Search path exhausted or match failed -> nothing found!
	   */

	  error = error_alloc();
	  error->error_type = QERR_nothing_found;
	  
	  error->baseobject = NULLCP;
	  error->ds_message = NULLCP;
	  
	  return error;
	}
      else
	/*
	 * Attempt a search for the target
	 */
	{
	  ufsrec->trying_target_search = TRUE;

	  to_follow = ufsrec->to_follow;

	  if ((error = follow_up(ufsrec->target_data, to_follow, ufsrec))
	      == NULLError)
	    {
	      dn_list_free(&to_follow);
	      ufsrec->to_follow = NULLEntryList;
	    }
	  else
	    if (error->error_type == QERR_invalid_search_class)
	      error->error_type = QERR_nothing_found;

	  return error;
	}
    }
  else
    {
      QCardinal succeeded = 0;
      errorList sub_error;

      if ((to_follow = ufsrec->to_follow) == NULLEntryList)
	/*
	 *	Can't continue search!
	 */
	{
	  add_error_to_list(&error,
			    NULLCP,
			    QERR_nothing_found,
			    (struct DSError *) NULL,
			    NULLCP);
	  
	  return error;
	}

      for (curr_srch_data = ufsrec->context_data;
	   curr_srch_data != NULLSearchData;
	   curr_srch_data = curr_srch_data->next)
	if (curr_srch_data->used == FALSE)
	  {
	    if ((sub_error = follow_up(curr_srch_data, to_follow, ufsrec))
		!= NULLError)
	      {
		if (sub_error->error_type != QERR_invalid_search_class)
		  {
		    sub_error->next = error;
		    error = sub_error;
		  }
		else
		  error_list_free(&sub_error);
	      }
	    else
	      succeeded++;
	  }
	  
      if (succeeded == 0)
	/*
	 *	Can't continue search!
	 */
	{
	  add_error_to_list(&error,
			    NULLCP,
			    QERR_nothing_found,
			    (struct DSError *) NULL,
			    NULLCP);
	  
	  return error;
	}

      dn_list_free(&to_follow);
      ufsrec->to_follow = NULLEntryList;
    }
  
  return NULLError;
} /* process_ufsearch */



/*
 * - follow_up() -
 *
 *
 */
static errorList follow_up(follow_data, baselist, ufsrec)
     searchData follow_data;
     entryList baselist;
     ufsearchRec ufsrec;
{
  QCardinal succeeded = 0;
  errorList sub_error,
            errors = NULLError;
  entryList curr_base;
  searchClass this_srch_class;
  QCardinal class_errors = 0;
  attrTypeList curr_oc_type;

  this_srch_class = get_srch_class_of_name(follow_data->type_label);

  for (curr_base = baselist;
       curr_base != NULLEntryList;
       curr_base = curr_base->next)
    {
      for (curr_oc_type = this_srch_class->base_object_classes;
	   curr_oc_type == NULLAttrTypeList;
	   curr_oc_type = curr_oc_type->next)
	if (curr_oc_type->type == curr_base->object_class ||
	    qy_in_hierarchy
	    (name2oc(curr_base->object_class->oa_ot.ot_stroid),
	     name2oc(curr_oc_type->type->oa_ot.ot_stroid))
	    == TRUE)
	  break;

      if (curr_oc_type == NULLAttrTypeList)
	{
	  class_errors++;
	  continue;
	}

      if ((sub_error = do_directory_search(curr_base->string_dn,
					   follow_data,
					   this_srch_class,
					   ufsrec))
	  != NULLError)
	{
	  sub_error->next = errors;
	  errors = sub_error;
	}
      else
	{
	  succeeded++;
	  follow_data->attempted = TRUE;
	}
    }

  if (succeeded == 0)
    {
      if (errors != NULLError)
	;
      else if (class_errors > 0)
	add_error_to_list(&errors,
			  NULLCP,
			  QERR_invalid_search_class,
			  (struct DSError *) NULL,
			  NULLCP);
    }
  else
    if (errors != NULLError)
      error_list_free(&errors);
  
  return errors;
} /* follow_up */


/*
 * - process_ufs_ds_result() -
 *
 *
 */
request_state process_ufs_ds_result(request, task_id, ds_result)
     requestRec request;
     int task_id;
     struct DSResult *ds_result;
{
  DsTask task_rec;
  ufsearchRec ufsrec = request->UFSEARCH_REC;
  int *ufs_task_array;

  QBool is_exact_match_result;
  QCardinal aindex, hit_count;

  struct ds_search_result *search_result;
  searchData curr_data;

  task_rec = _get_task_of_id(task_id);
  
  /*
   * This shouldn't happen
   */

  if (task_rec == NULLDsTask)
    return RQ_processing;


  /*
   * Check if result comes from a search using exact or approx match.
   */

  for (aindex = 0, ufs_task_array = ufsrec->exact_task_ids;
       ufs_task_array[aindex] != task_id && aindex < MAX_Ufsearch_Tasks;
       aindex++)
    ;


  /*
   * Not an exact search, so must have been an approx search.
   */

  if (aindex >= MAX_Ufsearch_Tasks)
    {
      for (aindex = 0, ufs_task_array = ufsrec->approx_task_ids;
           ufs_task_array[aindex] != task_id && aindex < MAX_Ufsearch_Tasks;
           aindex++)
        ;

      if (aindex < MAX_Ufsearch_Tasks)
        is_exact_match_result = FALSE;
    }
  else
    is_exact_match_result = TRUE;


  /*
   * If not an approx search, then something's wrong!
   */

  if (aindex >= MAX_Ufsearch_Tasks)
    {
      _task_complete(task_id);
      return RQ_processing;
    }

  
  /*
   * Now get the search results.
   */

  search_result = &ds_result->res_sr;
  correlate_search_results(search_result);

  search_result->srr_next = NULLSRR;

  hit_count = 0;

  if (search_result->CSR_entries != NULLENTRYINFO)
    {
      EntryInfo *entry_ptr;

      register stringCell curr_attr;
      attrValList got_attrs = NULLAVList, curr_av;
      searchData search_vals, curr_val_list;
      stringCell curr_val;

      AttributeType entry_oc_type;

      char *curr_name;
      QBool good_match;

      void get_read_attrs();

      search_vals = ufsrec->trying_target_search == TRUE?
        ufsrec->target_data:
        ufsrec->context_data;


      /*
       * Decode and add matching entry names
       */

      for (entry_ptr = search_result->CSR_entries;
           entry_ptr != NULLENTRYINFO;
           entry_ptr = entry_ptr->ent_next)
        {
          curr_name = qy_dn2str(entry_ptr->ent_dn);
          good_match = FALSE;

	  get_read_attrs(entry_ptr->ent_attr, &got_attrs, EDBOUT);

	  if (!is_exact_match_result)
            {
              if (got_attrs != NULLAVList)
                for (curr_av = got_attrs, good_match = FALSE;
                     curr_av != NULLAVList && good_match == FALSE;
                     curr_av = curr_av->next)
                  for (curr_attr = curr_av->val_list;
                       curr_attr != NULLStrCell && good_match == FALSE;
                       curr_attr = curr_attr->next)
                    for (curr_val_list = search_vals;
                         curr_val_list != NULLSearchData &&
                         good_match == FALSE;
                         curr_val_list = curr_val_list->next)
                      for (curr_val = curr_val_list->values;
                           curr_val != NULLStrCell;
                           curr_val = curr_val->next)
                        if (is_good_match(curr_val->string, curr_attr->string))
                          good_match = TRUE;
            }
	  else
	    good_match = TRUE;

	  
	  /*
	   *	Made a good match. Check that the entry's object class
	   *	matches any found in the search classes. If so add it to the
	   *	to follow list.
	   *
	   *	If result of target search add to to_follow list anyway
	   */

	  entry_oc_type = NULLAttrT;
	  
	  if (good_match == TRUE)
	    if ((ufsrec->trying_target_search == TRUE &&
		 is_target_class(got_attrs,
				 ufsrec->target_data->type_label)) ||
		(ufsrec->trying_target_search == FALSE &&
		 (entry_oc_type = check_get_entry_oc(got_attrs)) != NULLAttrT))
	      {
		(void) dn_list_add(curr_name,
				   &ufsrec->to_follow,
				   entry_oc_type);
		hit_count++;
		
		ufsrec->made_good_matches = TRUE;
	      }


	  if (curr_name != NULLCP)
	    (void) free(curr_name);
	  
	  if (got_attrs != NULLAVList)
	    {
              free_string_seq(&(got_attrs->attr_name));
              free_string_seq(&(got_attrs->val_list));

              (void) free((char *) got_attrs);
	      got_attrs = NULLAVList;
            }
	}
    }

  if (search_result->CSR_limitproblem != LSR_NOLIMITPROBLEM)
    {
      QE_error_code limitproblem;
      
      switch (search_result->CSR_limitproblem)
        {
        case LSR_TIMELIMITEXCEEDED:
          limitproblem = QERR_time_limit_reached;
          break;
	  
        case LSR_SIZELIMITEXCEEDED:
          limitproblem = QERR_size_limit_reached;
          break;

        case LSR_ADMINSIZEEXCEEDED:
          limitproblem = QERR_admin_limit_reached;
          break;
        }
      
      add_error_to_request_rec(request,
                               task_rec->baseobject,
                               limitproblem,
                               (struct DSError *) NULL,
                               NULLCP);
    }

#ifndef NO_STATS
  LLOG (log_stat, LLOG_NOTICE,
        ("SEARCH RESULT from +%s, task %d, %d good matches",
         task_rec->baseobject,
         task_id,
         hit_count));
#endif


  /*
   *	Flag appropriate search data as having been used if
   *	search has yielded results
   */

  if (hit_count > 0)
    {
      DsTask prev, curr = NULLDsTask;
      QBool found_task = FALSE;

      if (ufsrec->trying_target_search == TRUE)
	curr_data = ufsrec->target_data;
      else
	curr_data = ufsrec->context_data;
      
      while (curr_data != NULLSearchData && found_task == FALSE)
	{
	  for (curr = curr_data->task_list, prev = NULLDsTask;
	       curr != NULLDsTask;
	       curr = curr->next)
	    if (curr->task_id == task_id)
	      {
		curr_data->used = TRUE;
		
		if (prev == NULLDsTask)
		  curr_data->task_list = curr->next;
		else
		  prev->next = curr->next;

		ds_task_free(curr);

		found_task = TRUE;
		
		break;
	      }
	    else
	      prev = curr;

	  curr_data = curr_data->next;
	}
    }

  
  /*
   *	Remove task record
   */

  ufs_task_array[aindex] = NO_TASK;
	  
  if (is_exact_match_result == TRUE)
    ufsrec->exact_task_count--;
  else
    ufsrec->approx_task_count--;

  _task_complete(task_id);
  
  
  /*
   *	Check request status
   */

  return check_ufsearch_state(ufsrec);
} /* process_ufs_ds_result */

/*
 * - process_ufs_ds_result() -
 *
 *
 */
request_state process_ufs_ds_error(request, task_id, ds_error)
     requestRec request;
     int task_id;
     struct DSError *ds_error;
{
  DsTask task_rec;
  ufsearchRec ufsrec = request->UFSEARCH_REC;
  int *ufs_task_array;

  QBool is_exact_match_result;
  QCardinal aindex;


  task_rec = _get_task_of_id(task_id);
  
  /*
   * This shouldn't happen
   */

  if (task_rec == NULLDsTask)
    return RQ_processing;


  /*
   * Check if result comes from a search using exact or approx match.
   */

  for (aindex = 0, ufs_task_array = ufsrec->exact_task_ids;
       ufs_task_array[aindex] != task_id && aindex < MAX_Ufsearch_Tasks;
       aindex++)
    ;


  /*
   * Not an exact search, so must have been an approx search.
   */

  if (aindex >= MAX_Ufsearch_Tasks)
    {
      for (aindex = 0, ufs_task_array = ufsrec->approx_task_ids;
           ufs_task_array[aindex] != task_id && aindex < MAX_Ufsearch_Tasks;
           aindex++)
        ;

      if (aindex < MAX_Ufsearch_Tasks)
        is_exact_match_result = FALSE;
    }
  else
    is_exact_match_result = TRUE;


  /*
   * If not an approx search, then something's wrong!
   */

  if (aindex >= MAX_Ufsearch_Tasks)
    {
      _task_complete(task_id);
      return RQ_processing;
    }

  
  /*
   *  Log and save error
   */

  add_error_to_request_rec(request,
			   task_rec->baseobject,
			   get_log_error_type(ds_error, task_id),
			   ds_error,
			   NULLCP);

  /*
   *	Remove task record
   */

  ufs_task_array[aindex] = NO_TASK;
	  
  if (is_exact_match_result == TRUE)
    ufsrec->exact_task_count--;
  else
    ufsrec->approx_task_count--;

  _task_complete(task_id);
  
  
  /*
   *	Check request status
   */

  return check_ufsearch_state(ufsrec);
}

/*
 *  - do_directory_search() -
 *
 *
 */
static errorList do_directory_search(baseobject,
				     data,
				     srch_class,
				     ufsrec)
     char *baseobject;
     searchData data;
     searchClass srch_class;
     ufsearchRec ufsrec;
{
  stringCell values = data->values;

  struct ds_search_arg search_arg;
  struct DAPindication di;
  DN base_dn, rdn;
  
  int task_id;

  QCardinal rdn_comp_num, count;
  QBool has_org;

  attrTypeList match_attr;

  Filter exact, approx;
  Attr_Sequence get_attrs;
  
  errorList error = NULLError;
  
  DsTask new_task;


  /*
   *	Want object class back plus the match attributes
   */

  get_attrs = as_comp_new(AttrT_new("2.5.4.0"), NULLAV, NULLACL_INFO);

  for (match_attr = srch_class->search_attrs;
       match_attr != NULLAttrTypeList;
       match_attr = match_attr->next)
    get_attrs = as_merge(get_attrs,
			 as_comp_new(match_attr->type, NULLAV, NULLACL_INFO));
  
  
  /*
   *	Extract data from DN
   */
  
  if (baseobject == NULLCP || *baseobject == '\0')
    {
      base_dn = NULLDN;
      rdn_comp_num = 0;
      has_org = FALSE;
    }
  else
    {
      AttributeType org_type = AttrT_new("2.5.4.10");
      base_dn = str2dn(baseobject);

      for (rdn = base_dn->dn_parent, rdn_comp_num = 1;
	   rdn != NULLDN;
	   rdn = rdn->dn_parent, rdn_comp_num++)
	if (rdn->dn_rdn->rdn_at == org_type)
	  has_org = TRUE;
    }


  /*
   *	Standard args
   */

  (void) get_default_service(&search_arg.sra_common);

  search_arg.sra_common.ca_servicecontrol.svc_options =
    search_arg.sra_common.ca_servicecontrol.svc_options | SVC_OPT_PREFERCHAIN;

  search_arg.sra_common.ca_servicecontrol.svc_sizelimit = 10;
  
  search_arg.sra_eis.eis_allattributes = FALSE;
  search_arg.sra_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;
  search_arg.sra_eis.eis_select = get_attrs;
  
  search_arg.sra_baseobject = base_dn;

  if (rdn_comp_num >= 2 && has_org)
    search_arg.sra_subset = SRA_WHOLESUBTREE;
  else
    search_arg.sra_subset = SRA_ONELEVEL;


  if ((error = make_search_filters(values, srch_class, &exact, &approx))
      != NULLError)
    {
      if (base_dn != NULLDN)
	dn_free(base_dn);

      as_free(get_attrs);

      return error;
    }
  
  
  /*
   *	Record ds request invocations, then send of approx and exact
   *	search requests
   */

  for (count = 0;
       ufsrec->exact_task_ids[count] != NO_TASK && count < MAX_Ufsearch_Tasks;
       count++)
    ;
  
  if (count == MAX_Ufsearch_Tasks)
    {
      filter_free(exact);
      filter_free(approx);

      if (base_dn != NULLDN)
	dn_free(base_dn);
      
      as_free(get_attrs);
      
      add_error_to_list(&error,
			NULLCP,
			QERR_internal_limit_reached,
			(struct DSError *) NULL,
			NULLCP);

      return error;
    }
  

  if (_task_invoked(SEARCH_TASK, baseobject, ufsrec->request_id, &task_id)
      != QERR_ok)
    {
      filter_free(exact);
      filter_free(approx);

      if (base_dn != NULLDN)
	dn_free(base_dn);
      
      as_free(get_attrs);
      
      add_error_to_list(&error,
			NULLCP,
			QERR_internal_limit_reached,
			(struct DSError *) NULL,
			NULLCP);

      return error;
    }

  search_arg.sra_filter = exact;

  if (DapSearch(dsap_ad, task_id, &search_arg, &di, ROS_ASYNC) == NOTOK)
    {
      _task_complete(task_id);

      filter_free(exact);
      filter_free(approx);

      if (base_dn != NULLDN)
	dn_free(base_dn);
      
      as_free(get_attrs);
      
      add_error_to_list(&error,
			NULLCP,
			QERR_internal_limit_reached,
			(struct DSError *) NULL,
			NULLCP);

      return error;
    }

  ufsrec->exact_task_ids[count] = task_id;
  ufsrec->exact_task_count++;
  ufsrec->tasks_sent++;

  
  new_task = ds_task_alloc();
  
  new_task->baseobject = NULLCP;
  new_task->task_id = task_id;

  new_task->next = data->task_list;
  data->task_list = new_task;
  

  filter_free(exact);

#ifndef NO_STATS
  LLOG (log_stat, LLOG_NOTICE,
        ("EXACT MATCH +%s, task %d, extent %d value %s",
         baseobject,
	 task_id,
	 search_arg.sra_subset,
	 values->string));
#endif


  for (count = 0;
       ufsrec->approx_task_ids[count] != NO_TASK && count < MAX_Ufsearch_Tasks;
       count++)
    ;
  
  if (count == MAX_Ufsearch_Tasks)
    {
      filter_free(approx);

      if (base_dn != NULLDN)
	dn_free(base_dn);
      
      as_free(get_attrs);
      
      add_error_to_list(&error,
			NULLCP,
			QERR_internal_limit_reached,
			(struct DSError *) NULL,
			NULLCP);

      return error;
    }
  

  if (_task_invoked(SEARCH_TASK, baseobject, ufsrec->request_id, &task_id)
      != QERR_ok)
    {
      filter_free(approx);

      if (base_dn != NULLDN)
	dn_free(base_dn);
      
      as_free(get_attrs);
      
      add_error_to_list(&error,
			NULLCP,
			QERR_internal_limit_reached,
			(struct DSError *) NULL,
			NULLCP);

      return error;
    }

  search_arg.sra_filter = approx;

  if (DapSearch(dsap_ad, task_id, &search_arg, &di, ROS_ASYNC) == NOTOK)
    {
      _task_complete(task_id);

      filter_free(approx);

      if (base_dn != NULLDN)
	dn_free(base_dn);
      
      as_free(get_attrs);
      
      add_error_to_list(&error,
			NULLCP,
			QERR_internal_limit_reached,
			(struct DSError *) NULL,
			NULLCP);

      return error;
    }

  ufsrec->approx_task_ids[count] = task_id;
  ufsrec->approx_task_count++;
  ufsrec->tasks_sent++;

  new_task = ds_task_alloc();
  
  new_task->baseobject = NULLCP;
  new_task->task_id = task_id;

  new_task->next = data->task_list;
  data->task_list = new_task;


  filter_free(approx);

#ifndef NO_STATS
  LLOG (log_stat, LLOG_NOTICE,
        ("APPROX MATCH +%s, task %d, extent %d value %s",
         baseobject,
	 task_id,
	 search_arg.sra_subset,
	 values->string));
#endif

  if (base_dn != NULLDN)
    dn_free(base_dn);
      
  as_free(get_attrs);
  
  return NULLError;
} /* do_directory_search */

/*
 * - make_search_filters() -
 *
 *
 */
static errorList make_search_filters(values, srch_class, exact_ptr, approx_ptr)
     stringCell values;
     searchClass srch_class;
     Filter *exact_ptr, *approx_ptr;
{
  attrTypeList curr_attr;
  stringCell curr_value;

  AttributeType curr_type;

  Filter exact_or = NULLFILTER,
         approx_or = NULLFILTER,
         curr_exact = NULLFILTER,
         curr_approx = NULLFILTER,
  	 tmp_exact,
         tmp_approx;
 


  for (curr_attr = srch_class->search_attrs;
       curr_attr != NULLAttrTypeList;
       curr_attr = curr_attr->next)
    {
      curr_type = curr_attr->type;
      
      for (curr_value = values;
	   curr_value != NULLStrCell;
	   curr_value = curr_value->next)
	{
	  if (make_filter_items(curr_type,
				curr_value->string,
				&tmp_exact,
				&tmp_approx)
	      != QERR_ok)
	    {
	      errorList error;
	      char err_msg[LINESIZE];
	      
	      if (exact_or != NULLFILTER)
		{
		  filter_free(exact_or);
		  filter_free(approx_or);
		}

	      (void) sprintf(err_msg,
			     "Bad syntax in value `%s'!",
			     curr_value->string);


	      error = NULLError;

	      add_error_to_list(&error,
				NULLCP,
				QERR_bad_value_syntax,
				(struct DSError *) NULL,
				err_msg);
	      
	      return error;
	    }

	  if (exact_or == NULLFILTER)
	    {
	      exact_or = tmp_exact;
	      approx_or = tmp_approx;
	    }

	  if (curr_exact == NULLFILTER)
	    {
	      curr_exact = tmp_exact;
	      curr_approx = tmp_approx;
	    }
	  else
	    {
	      curr_exact = curr_exact->flt_next = tmp_exact;
	      curr_approx = curr_approx->flt_next = tmp_approx;
	    }
	}
    }

  /*
   *	A chain of approximate and exact filter items has been made.
   */

  curr_exact = filter_alloc();
  curr_approx = filter_alloc();

  *exact_ptr = curr_exact;
  *approx_ptr = curr_approx;

  curr_exact->flt_type = curr_approx->flt_type = FILTER_OR;
  curr_exact->flt_next = curr_approx->flt_next = NULLFILTER;

  curr_exact->FUFILT = exact_or;
  curr_approx->FUFILT = approx_or;

  return NULLError;
} /* make_search_filters */


/*
 * - check_ufsearch_state() -
 *
 *
 */
static request_state check_ufsearch_state(ufsrec)
     ufsearchRec ufsrec;
{
  ufsearchResult result;
  ufsearchReport report;
  entryList hits = ufsrec->to_follow;
  errorList error;
  requestRec request = _get_request_of_id(ufsrec->request_id);
      
  if (ufsrec->exact_task_count != 0 || ufsrec->approx_task_count != 0)
    return RQ_processing;
  else
    {
      result = ufs_result_alloc();
      
      result->tasks_sent = ufsrec->tasks_sent;
      result->tasks_failed = ufsrec->tasks_failed;
      
      if (ufsrec->trying_target_search == TRUE)
	{
	  if (hits == NULLEntryList)
	    /*
	     *	Search for target has failed, try next element of search path
	     */
	    {
	      if ((error = process_ufsearch(ufsrec)) == NULLError)
		{
		  (void) free((char *) result);

		  report = ufsrec->report = ufsearch_report_alloc();
		  
		  report->type = UFS_NewAttempt;
		  report->matched_list = NULLStrCell;

		  return RQ_report_returned;
		}
	      else
		{
		  result->hits = NULLEntryList;
		  result->hit_num = 0;
		  
		  result->errors = error;
		  error->next = request->errors;

		  request->errors = NULLError;
		  
		  result->search_status = UFS_Failed;
		  
		  ufsrec->result = result;
		  
		  return RQ_results_returned;
		}
	    }
	  else
	    /*
	     *	Found something!
	     */
	    {
	      result->hits = hits;
	      result->hit_num = 0;
	      
	      ufsrec->to_follow = NULLEntryList;
	      
	      result->errors = request->errors;
	      request->errors = NULLError;
	      
	      result->search_status = UFS_Succeeded;
	      
	      ufsrec->result = result;
	      
	      return RQ_results_returned;
	    }
	}
      else
	/*
	 *  The search result just processed was a context search
	 */
	{
	  if (hits == NULLEntryList)
	    {
	      /*
	       *  Search has hit a dead end, try search path
	       */
	      if ((error = process_ufsearch(ufsrec)) == NULLError)
		{
		  (void) free((char *) result);

		  report = ufsrec->report = ufsearch_report_alloc();

		  report->type = UFS_NewAttempt;
		  report->matched_list = NULLStrCell;
		  
		  return RQ_report_returned;
		}
	      else
		{
		  result->hits = NULLEntryList;
		  result->hit_num = 0;
		  
		  result->errors = error;
		  error->next = request->errors;

		  request->errors = NULLError;
		  
		  result->search_status = UFS_Failed;
		  
		  ufsrec->result = result;
		  
		  return RQ_results_returned;
		}
	    }

	  /*
	   *	Hits have been made, follow them up
	   */
	  
	  if ((error = process_ufsearch(ufsrec)) == NULLError)
	    {
	      searchData curr_data;

	      (void) free((char *) result);

	      report = ufsrec->report = ufsearch_report_alloc();
	      
	      report->type = UFS_MadeMatches;

	      report->matched_list = NULLStrCell;
	      for (curr_data = ufsrec->context_data;
		   curr_data != NULLSearchData;
		   curr_data = curr_data->next)
		if (curr_data->used == TRUE)
		  add_string_to_seq(curr_data->type_label,
				    &report->matched_list);
	      
	      return RQ_report_returned;
	    }
	  else
	    {
	      result->hits = NULLEntryList;
	      result->hit_num = 0;
	      
	      result->errors = error;
	      error->next = request->errors;
	      
	      request->errors = NULLError;
	      
	      result->search_status = UFS_Failed;
	      
	      ufsrec->result = result;
	      
	      return RQ_results_returned;
	    }
	}
    }
} /* check_ufsearch_state */


/*
 * - add_srch_class() -
 *
 *
 */
QBool add_search_class(class_name,
		       target_type,
		       search_attributes,
		       base_classes)
     char *class_name;
     AttributeType target_type;
     attrTypeList search_attributes, base_classes;
{
  QCardinal count = search_class_num;
  searchClass new_class;

  if (count == MAX_Search_Classes)
    return FALSE;

  new_class = search_classes[search_class_num++] = search_class_alloc();

  new_class->label = copy_string(class_name);
  new_class->search_attrs = search_attributes;

  new_class->target_object_class = target_type;
  new_class->base_object_classes = base_classes;

  return TRUE;
} /* add_srch_class */


/*
 * - get_srch_class_of_name() -
 *
 *
 */
searchClass get_srch_class_of_name(name)
     char *name;
{
  QCardinal count;
  
  for (count = 0; count < search_class_num; count++)
    if (lexequ(search_classes[count]->label, name) == 0)
      return search_classes[count];

  return NULLSearchClass;
} /* get_srch_class_of_name */


/*
 * - add_target_class() -
 *
 *
 */
QBool add_target_class(class_name, context_classes)
     char *class_name;
     stringCell context_classes;
{
  QCardinal count;
  targetClass new_class;
  
  
  if (target_class_num == MAX_Target_Classes)
    return FALSE;

  if (class_name == NULLCP || *class_name == '\0')
    return FALSE;

  
  for (count = 0; count < target_class_num; count++)
    if (lexequ(target_classes[count]->target_label, class_name) == 0)
      break;


  if (count == target_class_num)
    {
      new_class = target_classes[target_class_num++] = target_class_alloc();
      new_class->target_label = copy_string(class_name);
    }
  else
    new_class = target_classes[count];


  new_class->context_classes = context_classes;
  new_class->target_paths = NULLPathRule;

  return TRUE;
} /* add_target_class */


/*
 * - add_target_path() -
 *
 *
 */
void add_target_path(class_name, srch_class_list, path)
     char *class_name;
     stringCell srch_class_list;
     entryList path;
{
  QCardinal count;
  targetClass target;
  pathRule new_rule;
  
  for (count = 0; count < target_class_num; count++)
    if (lexequ(target_classes[count]->target_label, class_name) == 0)
      break;

  if (count == target_class_num)
    (void) add_target_class(class_name, NULLStrCell);

  
  target = target_classes[count];

  new_rule = path_rule_alloc();

  new_rule->next = target->target_paths;
  target->target_paths = new_rule;

  new_rule->need_classes = srch_class_list;
  new_rule->default_path = path;
}


/*
 * - get_target_class() -
 *
 *
 */
targetClass get_target_class(class_name)
     char *class_name;
{
  QCardinal count;
  
  for (count = 0; count < target_class_num; count++)
    if (lexequ(target_classes[count]->target_label, class_name) == 0)
      break;

  if (count == target_class_num)
    return NULLTargetClass;
  
  return target_classes[count];
}

/*
 * - is_target_class() -
 *
 *
 */
static QBool is_target_class(attrs, target_label)
     attrValList attrs;
     char *target_label;
{
  searchClass target = get_srch_class_of_name(target_label);
  attrValList curr_av;
  AttributeType oclass_type = AttrT_new("2.5.4.0");
  stringCell oc_vals, curr_str_oc;
  objectclass *target_oc;

  if (target == NULLSearchClass)
    return FALSE;

  target_oc = name2oc(target->target_object_class->oa_ot.ot_stroid);

  for (curr_av = attrs; curr_av != NULLAVList; curr_av = curr_av->next)
    if (AttrT_new(curr_av->attr_name->string) == oclass_type)
      break;

  if (curr_av == NULLAVList)
    return FALSE;

  oc_vals = make_sequence_from_values(curr_av->val_list);

  for (curr_str_oc = oc_vals;
       curr_str_oc != NULLStrCell;
       curr_str_oc = curr_str_oc->next)
    if (target_oc == name2oc(curr_str_oc->string))
      {
	free_string_seq(&oc_vals);
	return TRUE;
      }
    
  
  free_string_seq(&oc_vals);

  return FALSE;
}

/*
 * - check_get_entry_oc() -
 *
 *
 */
static AttributeType check_get_entry_oc(attrs)
     attrValList attrs;
{
  AttributeType oclass_type = AttrT_new("2.5.4.0");
  attrValList curr_av;
  attrTypeList parent_ocs = NULLAttrTypeList,
  	       curr_parent1,
               curr_parent2,
               curr_oc_type;
  stringCell curr_val, oc_vals = NULLStrCell;
  objectclass *curr_oc;
  QCardinal count;


  for (curr_av = attrs; curr_av != NULLAVList; curr_av = curr_av->next)
    if (AttrT_new(curr_av->attr_name->string) == oclass_type)
      break;

  if (curr_av == NULLAVList)
    return NULLAttrT;

  oc_vals = make_sequence_from_values(curr_av->val_list);

  for (curr_val = oc_vals; curr_val != NULLStrCell; curr_val = curr_val->next)
    {
      curr_oc = name2oc(curr_val->string);
      
      for (count = 0; count < search_class_num; count++)
	for (curr_oc_type = search_classes[count]->base_object_classes;
	     curr_oc_type != NULLAttrTypeList;
	     curr_oc_type = curr_oc_type->next)
	  if (curr_oc_type->type == NULLAttrT)
	    continue;
	  else if (qy_in_hierarchy
		   (name2oc(curr_oc_type->type->oa_ot.ot_stroid),
		    curr_oc)
		   == TRUE)
	    (void) add_type_to_list((AttributeType) curr_oc, &parent_ocs);
    }

  if (parent_ocs == NULLAttrTypeList)
    return NULLAttrT;
  else if (parent_ocs->next == NULLAttrTypeList)
    curr_oc = (objectclass *) parent_ocs->type;
  else
    {
      for (curr_parent1 = parent_ocs;
	   curr_parent1 != NULLAttrTypeList;
	   curr_parent1 = curr_parent1->next)
	{
	  curr_oc = (objectclass *) curr_parent1->type;

	  for (curr_parent2 = parent_ocs;
	       curr_parent2 != NULLAttrTypeList;
	       curr_parent2 = curr_parent1->next)
	    if (qy_in_hierarchy((objectclass *) curr_parent2->type,
				curr_oc)
		== TRUE)
	      break;

	  if (curr_parent2 != NULLAttrTypeList)
	    break;
	}
    }

  free_at_list(&parent_ocs);

  return AttrT_new(curr_oc->oc_ot.ot_stroid);
} /* check_get_entry_oc */


/*
 * - get_default_path() -
 *
 *
 */
static QBool get_default_path(target_data, context_data, return_path_ptr)
     searchData target_data, context_data;
     entryList *return_path_ptr;
{
  searchData curr_data;
  stringCell curr_class_name;
  pathRule curr_path_rule;
  targetClass target;
  QCardinal count;

  *return_path_ptr = NULLEntryList;


  for (count = 0; count < target_class_num; count++)
    if (lexequ(target_classes[count]->target_label, target_data->type_label)
	== 0)
      break;


  if (count == target_class_num)
    {
      (void) dn_list_insert("", return_path_ptr, NULLAttrT);
      return TRUE;
    }
  
  
  target = target_classes[count];
  
  for (curr_path_rule = target->target_paths;
       curr_path_rule != NULLPathRule;
       curr_path_rule = curr_path_rule->next)
    {
      if (context_data == NULLSearchData)
	{
	  if (curr_path_rule->need_classes == NULLStrCell ||
	      (*curr_path_rule->need_classes->string == '\0' &&
	       curr_path_rule->need_classes->next == NULLStrCell))
	    {
	      dn_list_copy(curr_path_rule->default_path, return_path_ptr);
	      return TRUE;
	    }
	  else
	    continue;
	}

      for (curr_data = context_data;
	   curr_data != NULLSearchData;
	   curr_data = curr_data->next)
	{
	  for (curr_class_name = curr_path_rule->need_classes;
	       curr_class_name != NULLStrCell;
	       curr_class_name = curr_class_name->next)
	    if (lexequ(curr_class_name->string, curr_data->type_label) == 0)
	      break;

	  if (curr_class_name == NULLStrCell)
	    break;
	}
      
      if (curr_data == NULLSearchData)
	break;
    }

  if (curr_path_rule == NULLPathRule)
    (void) dn_list_insert("", return_path_ptr, NULLAttrT);
  else if (curr_path_rule->default_path == NULLEntryList)
    (void) dn_list_insert("", return_path_ptr, NULLAttrT);
  else
    dn_list_copy(curr_path_rule->default_path, return_path_ptr);
  

  return TRUE;
} /* get_default_path */


/*
 * Freeing procedures
 */

void ufsearch_rec_free(ufsrec)
     ufsearchRec ufsrec;
{
  if (ufsrec->to_follow != NULLEntryList)
    dn_list_free(&ufsrec->to_follow);

  if (ufsrec->default_path != NULLEntryList)
    dn_list_free(&ufsrec->default_path);

  search_data_free(&ufsrec->context_data);
  search_data_free(&ufsrec->target_data);

  (void) free((char *) ufsrec);
}

void ufsearch_report_free(report_ptr)
     ufsearchReport *report_ptr;
{
  ufsearchReport report = *report_ptr;
  
  if (report == NULLUfsearchReport)
    return;

  free_string_seq(&report->matched_list);
  
  free((char *) report);

  *report_ptr = NULLUfsearchReport;
}

void ufsearch_result_free(result_ptr)
     ufsearchResult *result_ptr;
{
  ufsearchResult result = *result_ptr;

  if (result == NULLUfsearchResult)
    return;

  if (result->hits != NULLEntryList)
    dn_list_free(&result->hits);
  
  if (result->errors != NULLError)
    error_list_free(&result->errors);
  
  (void) free((char *) result);
  
  *result_ptr = NULLUfsearchResult;
}

void search_data_free(list_ptr)
     searchData *list_ptr;
{
  searchData list = *list_ptr, next;

  if (list == NULLSearchData)
    return;

  while (list != NULLSearchData)
    {
      next = list->next;

      free_string_seq(&list->values);
      ds_task_list_free(&list->task_list);
      
      (void) free((char *) list);
      
      list = next;
    }
}


/*
 * Misc procs
 *
 *
 */
static stringCell make_sequence_from_values(strval_list)
     stringCell strval_list;
{
  stringCell strval, new_vals = NULLStrCell;
  register char *start, *end, save;

  for (strval = strval_list; strval != NULLStrCell; strval = strval->next)
    {
      start = end = strval->string;
      
      while (*end != '\0' && *end != '\n')
	{
	  start = end;
	  
	  while (isspace(*start))
	    start++;
	  
	  end = start;
	  
	  while (!isspace(*end) && *end != '\n')
	    end++;
	  save = *end;

	  *end = '\0';

	  add_string_to_seq(start, &new_vals);

	  *end = save;
	}
    }
  
  return new_vals;
}
