%{
#include "layout.h"
#include "configure.h"
#include "defs.h"
#include "actions.h"
#include "dn_list.h"
#include "query.h"
#include "sequence.h"

int yylex();
void exit();

extern FILE *configuration_file;
extern Cardinal config_line_num;
extern char config_file_path[];
extern readAttribute attributes_to_display;
extern String default_target;
extern char bitmap_directory[], help_directory[];

static char lexeme[255];

%}
%start configure

%union
{
  char strval[255];
  Cardinal symbol;
  layoutMember layout;
  orientation orient;
  outputMode mode;
  objectResource resource;
  translationList trans;
  actionType atype;
  stringCell string_list;
  entryList entry_list;
  int integer;
}

%token STATUS TITLE CURR_POS READ_OUT LOOKUP ENTRY_LIST HELP MODIFY
%token MOVETO BUTTON ERROR_VIEW
%token HORIZONTAL VERTICAL START ACTION SETACTION SETBIND BITMAP HELPFILE
%token WIDTH HEIGHT EXPAND INSENSITIVE LABEL NAME
%token HELP_DIR BITMAP_DIR
%token A_RAISEOBJ A_KEEPOBJ A_ABORT A_CLOSEOBJ A_QUITAPP A_ENTRYSEL A_ACTIVATE
%token MODIFY_TYPE ATTR_IMPLIES ATTR_TEMPLATE
%token UFN_PATH
%token UFS_CLASS UFS_TARGET UFS_DEFAULTS DEFAULT_TARGET

%token <symbol> STATUS TITLE CURR_POS READ_OUT LOOKUP ENTRY_LIST HELP
%token <symbol> MOVETO BUTTON MODIFY ERROR_VIEW
%token <symbol> GROUP LAYOUT 
%token <symbol> ':' '(' ')'
%token <symbol> HORIZONTAL VERTICAL START
%token <symbol> NEW SINGLE GLOBALMANAGED LOCALMANAGED CURRENT
%token <symbol> A_RAISEOBJ A_KEEPOBJ A_ABORT A_CLOSEOBJ A_QUITAPP A_ENTRYSEL
%token <symbol> MODIFY_TYPE ATTR_IMPLIES ATTR_TEMPLATE
%token <symbol> READ_TYPE ATTR_NAME UFN_PATH
%token UFS_CLASS UFS_TARGET UFS_DEFAULTS DEFAULT_TARGET

%token <strval> STRING 

%type <layout> 		layout_def layout_list
%type <symbol> 		primitive_window
%type <mode> 		output_mode
%type <orient> 		orient
%type <resource> 	option option_list prim_options layout_options
%type <trans> 		trans_list trans_def activateable
%type <string_list> 	string_list
%type <entry_list>	entry_list oc_entry_list
%type <integer> 	ufn_bound

%%
/*
 * 	Main rule.
 */
 configure:	var_list
      	   	;

/*
 *	The configure options
 */

 var_list:	LAYOUT ':' STRING ':' layout_def  var_list
		  {
			put_layout(add_template_entry($3), $5);
		  }

		| START ':' start_list var_list		{;}

		| ACTION ':' action_def var_list	{;}

                | modify_option	var_list		{;}

		| read_option var_list			{;}
	
                | attr_name_option var_list		{;}

		| ufn_path_option var_list		{;}
	
		| ufs_path_option var_list		{;}
	
		| ufs_target_class_option var_list	{;}
	
		| ufs_cls_option var_list		{;}

		| ufs_default_option var_list		{;}
	
		| bitmaps_option var_list		{;}
	
		| helps_option var_list			{;}
	
		|					{;}
		;

/*
 *	Define directories
 */

 bitmaps_option:	BITMAP_DIR ':' STRING
			  {
			    (void) strcpy(bitmap_directory, $3);
			  }

 helps_option:		HELP_DIR ':' STRING
			  {
			    (void) strcpy(help_directory, $3);
			  }

/*
 *	Rule defining list of windows to appear on startup
 */

 start_list:   	STRING ',' start_list
		  {
		 	add_to_startup_list(add_template_entry($1));
		  }

		| STRING
		  {
			add_to_startup_list(add_template_entry($1));
		  }
                ;

/*
 *	Layout definition rules for  an XLookup window
 */

 layout_def:	orient '{' layout_list
		  {
			$3->orient = $1; $$ = $3;
		  }
  		;


 orient:	HORIZONTAL	{$$ = horizontal;}

		| VERTICAL 	{$$ = vertical;}
		;


 layout_list:	STRING layout_options layout_list
	 	  {
		       $$ = add_branch_to_layout(add_template_entry($1),$2,$3);
		  }

		| primitive_window prim_options layout_list
		  {
		  	$$ = add_prim_to_layout($1, $2, $3);
		  }

  		| '}'
		  {$$ = NULLLayout;}
		;

/*
 *	The flags and variables associated with window objects
 */

 layout_options: '(' option_list  		     {$$ = $2;}

		|				     {$$ = NULLObjRes;}
		;


 prim_options:	'(' option_list  		     {$$ = $2;}

		| 				     {$$ = NULLObjRes;}
		;

 option_list:	'(' option ')' option_list 	     {$2->next = $4; $$ = $2;}

		| ')'				     {$$ = NULLObjRes;}
		;


 option:	SETACTION STRING
		 {
		   $$ = put_resource(OT_action, (XtPointer) $2);
		 }

		| BITMAP STRING
		  {
		    $$ = put_resource(OT_bitmap, (XtPointer) $2);
		  }

		| HELPFILE STRING
		  {
		    $$ = put_resource(OT_helpfile, (XtPointer) $2);
		  }

		| HEIGHT STRING
		  {
		    $$ = put_resource(OT_height, (XtPointer) $2);
		  }

		| WIDTH STRING 
		  {
		    $$ = put_resource(OT_width, (XtPointer) $2);
		  }

		| EXPAND
		  {
		    $$ = put_resource(OT_expand, (XtPointer) NULL);
		  }
	
		| INSENSITIVE '(' string_list ')'
		  {
		    $$ = put_resource(OT_insensitive, (XtPointer) $3);
		  }

		| LABEL STRING
		  {
		    $$ = put_resource(OT_label, (XtPointer) $2);
		  }

		| NAME STRING
		  {
		    $$ = put_resource(OT_name, (XtPointer) $2);
		  }
		;


/*
 *	Rules for actions definitions
 */

 action_def:    STRING ':' '(' trans_list
		  {
			add_action($1, $4);
		  }
                ;


 trans_list:	trans_def ',' trans_list 	{$1->next = $3; $$ = $1;}

		| trans_def ')'   		{$$ = $1;}
		;			


 trans_def:	A_ABORT '(' ')'
                  {
		    $$ = new_trans(AT_AbortRequest,
				   NULLString,
				   NULLString,
				   NULLString,
				   CurrentWin,
				   (XtPointer) NULL);
		  }
  

		| A_RAISEOBJ '(' STRING ',' output_mode ')'
		  {
		    $$ = new_trans(AT_RaiseObject,
				   NULLString,
				   $3,
				   NULLString,
				   $5,
				   (XtPointer) NULL);
		  }

		| A_KEEPOBJ '(' ')'
		  {
		    $$ = new_trans(AT_KeepObject,
				   NULLString,
				   NULLString,
				   NULLString,
				   CurrentWin,
				   (XtPointer) NULL);
		  }

		| A_CLOSEOBJ '(' ')'
		  {
		    $$ = new_trans(AT_CloseObject,
				   NULLString,
				   NULLString,
				   NULLString,
				   CurrentWin,
				   (XtPointer) NULL);
		  }

		| A_QUITAPP '(' ')'
		  {
		    $$ = new_trans(AT_QuitApp,
				   NULLString,
				   NULLString,
				   NULLString,
				   CurrentWin,
				   (XtPointer) NULL);
		  }

		| A_ENTRYSEL '(' STRING ',' output_mode ')'
		  {
		    $$ = new_trans(AT_EntrySelect,
				   NULLString,
				   $3,
				   NULLString,
				   $5,
				   (XtPointer) NULL);
		  }

		| A_ACTIVATE '(' activateable ')'
		  {
		    $$ = $3;
		  }
		;


 activateable: 	LOOKUP ',' STRING ',' STRING ',' STRING ',' output_mode
  		 {
		   $$ = new_trans(AT_Activate,$3,$5,$7,$9, (XtPointer) LOOKUP);
		 }

		| MOVETO ',' STRING ',' STRING ',' STRING ',' output_mode
  		 {
		   $$ = new_trans(AT_Activate,$3,$5,$7,$9, (XtPointer) MOVETO);
		 }

		| MODIFY ',' STRING ',' STRING ',' STRING ',' output_mode
  		 {
		   $$ = new_trans(AT_Activate,$3,$5,$7,$9, (XtPointer) MODIFY);
		 }

		| READ_OUT ',' STRING ',' STRING ',' STRING ',' output_mode
  		 {
		   $$ = new_trans(AT_Activate,$3,$5,$7,$9,(XtPointer)READ_OUT);
		 }
		;

			     
 output_mode:   NEW		{$$ = NewAlways;}

                | SINGLE	{$$ = NewWhenNone;}

		| GLOBALMANAGED	{$$ = NewWhenKeptGlobal;}

		| GLOBALMANAGED	{$$ = NewWhenKeptLocal;}

		| CURRENT	{$$ = CurrentWin;}
		;
	
/*
 *	Defintions for primitive objects
 */

 primitive_window:	BUTTON		{$$ = $1;}
		   
  		   	| CURR_POS	{$$ = $1;}

			| ENTRY_LIST	{$$ = $1;}

			| ERROR_VIEW 	{$$ = $1;}

		   	| HELP 		{$$ = $1;}

			| LOOKUP	{$$ = $1;}

		   	| MOVETO	{$$ = $1;}

		   	| MODIFY	{$$ = $1;}

		   	| READ_OUT	{$$ = $1;}

		   	| STATUS	{$$ = $1;}

  		   	| TITLE		{$$ = $1;}
		   ;

/*
 *	Directory configuration rules
 */

/*
 *	UFN paths
 */

 ufn_path_option:	UFN_PATH ':' ufn_bound ',' ufn_bound ':' entry_list
			  {
			    add_ufn_path_element($3, $5, $7);
			  }

			| UFN_PATH ':' ufn_bound ':' entry_list
			  {
			    add_ufn_path_element($3, UfnNobound, $5);
			  }
			;

 ufn_bound:	STRING 
		  {
		    if (!isdigit(*($1)))
		      $$ = UfnNobound;
		    else
		      {
			int c;

			c = (int) strtol((char *) $1, (char **) NULL, 10);
			$$ = c;
		      }
		  }

		| '+'
 	  	  {
		    $$ = UfnPlus;
		  }

		;

 entry_list:	STRING
		  {
		    $$ = NULLEntryList;
		    (void) dn_list_insert($1, &($$), NULLAttrT);
		  }

		| STRING ',' entry_list 
		  {
		    $$ = $3;
		    (void) dn_list_insert($1, &($$), NULLAttrT);
		  }

		;

 oc_entry_list: STRING STRING 
                  {
		    AttributeType oc_type;

		    if ((oc_type = AttrT_new(name2oc($2)->oc_ot.ot_stroid))
			== NULLAttrT)
		      {
			(void) fprintf (stderr,
				      "\"%s\" - Error at or before line %d \n",
					config_file_path,
					config_line_num);

			(void) fprintf (stderr,
					"Invalid objectclass `%s'.Exiting.\n",
					$2);

			exit (1);
		      }

		    $$ = NULLEntryList;
		    (void) dn_list_insert($1, &($$), oc_type);
		  }

                | STRING STRING ',' oc_entry_list 
		  {
		    AttributeType oc_type;

		    if ((oc_type = AttrT_new(name2oc($2)->oc_ot.ot_stroid))
			== NULLAttrT)
		      {
			(void) fprintf (stderr,
				      "\"%s\" - Error at or before line %d \n",
					config_file_path,
					config_line_num);

			(void) fprintf (stderr,
					"Invalid objectclass `%s'.Exiting.\n",
					$2);

			exit (1);
		      }

		    $$ = $4;
		    (void) dn_list_insert($1, &($$), oc_type);
		  }

		; 


/*
 *	Form search paths
 */

 ufs_default_option: DEFAULT_TARGET ':' STRING
		       {
			 default_target = copy_string($3);
		       }

 ufs_path_option: UFS_DEFAULTS ':' STRING ':' string_list ':' oc_entry_list
		    {
		      add_target_path($3, $5, $7);
		    }
		  | UFS_DEFAULTS ':' STRING ':' string_list ':' 
		    {
		      add_target_path($3, $5, NULLEntryList);
		    }
		  | UFS_DEFAULTS ':' STRING ':' ':' oc_entry_list
		    {
		      add_target_path($3, NULLStrCell, $6);
		    }
		  | UFS_DEFAULTS ':' STRING ':' ':'
		    {
		      add_target_path($3, NULLStrCell, NULLEntryList);
		    }
	
 ufs_target_class_option: UFS_TARGET ':' STRING ':' string_list
                           {
			     if (add_target_class($3, $5) == FALSE)
			       {
				 (void) fprintf (stderr,
						"\"%s\" - Error at line %d \n",
						 config_file_path,
						 config_line_num);

				 (void) fprintf(stderr,
						"Maximum number of allowed ");

				 (void) fprintf(stderr,
						"searchTargets exceeded.");
				 
				 (void) fprintf(stderr, "Exiting.\n");
				 
				 exit (1);
			       }
			   }

	
 ufs_cls_option: UFS_CLASS ':' STRING ':' STRING ':' string_list ':' string_list
                   {
		     add_ufs_search_class($3, $5, $7, $9);
		   }

/*
 *	Read types
 */

 read_option:		READ_TYPE ':' read_attr_list {;}

 read_attr_list:	STRING read_attr_chain
			  {
			    add_display_attr($1, ShowAll);
			  }
			| '-' STRING read_attr_chain
			  {
			    add_display_attr($2, ShowLongest);
			  }
			|
			;

 read_attr_chain:	',' STRING read_attr_chain
			  {
			    add_display_attr($2, ShowAll);
			  }
			| ',' '-' STRING read_attr_chain
			  {
			    add_display_attr($3, ShowLongest);
			  }
			|
			;

/*
 *	User defined attribute names
 */
 attr_name_option:	ATTR_NAME ':' attr_name_list {;}
	
 attr_name_list:	STRING STRING attr_name_list
			  {
			    add_attribute_name($1, $2);
			  }

			| ',' STRING STRING attr_name_list
			  {
			    add_attribute_name($2, $3);
			  }
			
			|
			  {;}
			;


/*
 *	Modify configuration options
 */

 modify_option:		MODIFY_TYPE ':' modify_type_args	{;}

			| ATTR_IMPLIES ':' attr_implies_args	{;}

			| ATTR_TEMPLATE ':' attr_template_args	{;}
			;


 modify_type_args: STRING ':' STRING ':' STRING ':' string_list ':' string_list
		     {
		       set_entry_class($1, $3, $5, $7, $9);
		     }
		   | STRING ':' STRING ':' STRING ':' string_list ':'
                     {
		       set_entry_class($1, $3, $5, $7, NULLStrCell);
		     }
                   ;

  attr_implies_args:	STRING ':' STRING
		  	  {
		    	    set_attr_oc_rule($1, $3);
		  	  }
			;

 attr_template_args:	string_list ':' STRING ':' string_list
			  {
			    set_attr_template($3, $1, $5);
			  }
			;

/*
 *	Miscellaneous rules
 */

 string_list:	STRING
		  {
		    $$ = NULLStrCell;
		    (void) add_string_to_seq($1, &($$));
		  }

		| string_list ',' STRING 
		  {
		    $$ = $1;
		    (void) add_string_to_seq($3, &($$));
		  }
		;

%%

/*
 *	The lexer.
 */

int yylex()
{
  register Cardinal c;
  Cardinal count = 0;
  FILE *lexfile = configuration_file;
  register String lexptr;
  Boolean space;
  
  lexeme[0] = '\0';
  
  while (isspace((char) (c = getc(lexfile))))
    {
      if (c == EOF) return 0;
      if (c == '\n') config_line_num++;
    }
  
  lexeme[count++] = c;
  
  switch (c)
    {
    case '#':

      while ((c = getc(lexfile)) != '\n' && c != EOF)
	;

      config_line_num++;
      
      return yylex();

    case ':':
    case '(':
    case ')':
    case ',':
    case '{':
    case '}':
    case '-':
    case '+':

      lexeme[count] = '\0';
      return c;

    case '\"':

      count = 0;
      space = FALSE;

      while ((c = getc(lexfile)) != EOF && c != '\"')
	{
	  if (space == FALSE)
	    {
	      lexeme[count++] = (char) c;
	      
	      if (isspace((char) c))
		space = TRUE;
	    }
	  else if (!isspace((char) c))
	    {
	      lexeme[count++] = (char) c;
	      space = FALSE;
	    }
	}
    
      if (c == '\"')
	{
	  lexeme[count] = '\0'; 

	  (void) strcpy(yylval.strval, lexeme);
	  return STRING;
	} else
	  return 0;
    }

  while ((c = getc(lexfile)) != EOF && 
	 ! (c == '('
	    || c == ')'
	    || c == ':'
	    || c == '{'
	    || c == '}'
	    || c == ','
	    || c == '\"'
	    || isspace((char) c)))
    lexeme[count++] = (char) c;
  
  if (c == EOF)
    return 0;
  
  (void) fseek(lexfile, (long) -1, 1);
  
  lexeme[count] = '\0';
  
  if ((lexeme[0] == 'a' || lexeme[0] == 'A') && lexeme[1] == '.')
    {
      lexptr = (String) (lexeme + 2);

      switch (*lexptr)
	{
	case 'a':
	case 'A':

	  if (!lexequ("Activate", lexptr))
	    {
	      yylval.symbol = A_ACTIVATE;
	      return A_ACTIVATE;
	    }

	  if (!lexequ("AbortRequest", lexptr))
	    {
	      yylval.symbol = A_ABORT;
	      return A_ABORT;
	    }
      
	  break;
	  
	case 'c':
	case 'C':

	  if (!lexequ("CloseObject", lexptr))
	    {
	      yylval.symbol = A_CLOSEOBJ;
	      return A_CLOSEOBJ;
	    }
	  
	  break;
	  
	case 'e':
	case 'E':

	  if (!lexequ("EntrySelect", lexptr))
	    {
	      yylval.symbol = A_ENTRYSEL;
	      return A_ENTRYSEL;
	    }
	  
	  break;

	case 'k':
	case 'K':

	  if (!lexequ("KeepObject", lexptr))
	    {
	      yylval.symbol = A_KEEPOBJ;
	      return A_KEEPOBJ;
	    }
	  
	  break;

	case 'q':
	case 'Q':

	  if (!lexequ("QuitApp", lexptr))
	    {
	      yylval.symbol = A_QUITAPP;
	      return A_QUITAPP;
	    }
	  
	  break;
	  
	case 'r':
	case 'R':

	  if (!lexequ("RaiseObject", lexptr))
	    {
	      yylval.symbol = A_RAISEOBJ;
	      return A_RAISEOBJ;
	    }
	  
	  break;

	default:

	  (void) strcpy(yylval.strval, lexeme);
	  return STRING;
	}

      (void) strcpy(yylval.strval, lexeme);
      return STRING;
    }
  else if ((lexeme[0] == 'l' || lexeme[0] == 'L') && lexeme[1] == '.')
    {
      lexptr = (String) (lexeme + 2);
      
      switch (*lexptr)
	{
	case 'b':
	case 'B':
	  
	  if (!lexequ("Button", lexptr))
	    {
	      yylval.symbol = BUTTON;
	      return BUTTON;
	    }
	  
	  break;

	case 'e':
	case 'E':
	  
	  if (!lexequ("EntryList", lexptr))
	    {
	      yylval.symbol = ENTRY_LIST;
	      return ENTRY_LIST;
	    }

	  if (!lexequ("ErrorView", lexptr))
	    {
	      yylval.symbol = ERROR_VIEW;
	      return ERROR_VIEW;
	    }
	  
	  break;

	case 'h':
	case 'H':
	  
	  if (!lexequ("HelpWindow", lexptr))
	    {
	      yylval.symbol = HELP;
	      return HELP;
	    }
	  
	  break;

	case 'l':
	case 'L':
	  
	  if (!lexequ("lookup", lexptr))
	    {
	      yylval.symbol = LOOKUP;
	      return LOOKUP;
	    }
	  
	  break;

	case 'm':
	case 'M':
	  
	  if (!lexequ("ModifyEntry", lexptr))
	    {
	      yylval.symbol = MODIFY;
	      return MODIFY;
	    }
	  
	  if (!lexequ("MovetoEntry", lexptr))
	    {
	      yylval.symbol = MOVETO;
	      return MOVETO;
	    }
	  
	  break;

	case 'r':
	case 'R':
	  
	  if (!lexequ("ReadDisplay", lexptr))
	    {
	      yylval.symbol = READ_OUT;
	      return READ_OUT;
	    }
	  
	  break;

	case 's':
	case 'S':
	  
	  if (!lexequ("StatusBar", lexptr))
	    {
	      yylval.symbol = STATUS;
	      return STATUS;
	    }
	  
	  break;

	case 't':
	case 'T':
	  
	  if (!lexequ("TitleBar", lexptr))
	    {
	      yylval.symbol = TITLE;
	      return TITLE;
	    }
	  
	  break;

	default:
	  
	  (void) strcpy(yylval.strval, lexeme);
	  return STRING;
	}

      (void) strcpy(yylval.strval, lexeme);
      return STRING;

    }
  else
    {
      lexptr = (String) lexeme;

      switch (*lexptr)
	{
	case 'A':
	case 'a':

	  if (!lexequ("action", lexptr))
	    {
	      yylval.symbol = ACTION;
	      return ACTION;
	    }

	  if (!lexequ("attrOc", lexptr))
	    {
	      yylval.symbol = ATTR_IMPLIES;
	      return ATTR_IMPLIES;
	    }

	  if (!lexequ("attrValTemplate", lexptr))
	    {
	      yylval.symbol = ATTR_TEMPLATE;
	      return ATTR_TEMPLATE;
	    }

	  if (!lexequ("attrName", lexptr))
	    {
	      yylval.symbol = ATTR_NAME;
	      return ATTR_NAME;
	    }

	  break;
	  
	case 'b':
	case 'B':

	  if (!lexequ("bitmapDirectory", lexptr))
	    {
	      yylval.symbol = BITMAP_DIR;
	      return BITMAP_DIR;
	    }

	  if (!lexequ("bitmap", lexptr))
	    {
	      yylval.symbol = BITMAP;
	      return BITMAP;
	    }
	  
	  break;

	case 'c':
	case 'C':

	  if (!lexequ("current", lexptr))
	    {
	      yylval.symbol = CURRENT;
	      return CURRENT;
	    }
	
	  break;

	case 'd':
	case 'D':

	  if (!lexequ("DisplayAttribute", lexptr))
	    {
	      yylval.symbol = READ_TYPE;
	      return READ_TYPE;
	    }

	  if (!lexequ("DefaultTarget", lexptr))
	    {
	      yylval.symbol = DEFAULT_TARGET;
	      return DEFAULT_TARGET;
	    }

	  break;
	  
	case 'e':
	case 'E':

	  if (!lexequ("expand", lexptr))
	    {
	      yylval.symbol = EXPAND;
	      return EXPAND;
	    }
	  
	  if (!lexequ("entryClass", lexptr))
	    {
	      yylval.symbol = MODIFY_TYPE;
	      return MODIFY_TYPE;
	    }

	  break;

	case 'G':
	case 'g':

	  if (!lexequ("globalmanaged", lexptr))
	    {
	      yylval.symbol = GLOBALMANAGED;
	      return GLOBALMANAGED;
	    }
	  
	  break;
	  
       	case 'h':
	case 'H':

	  if (!lexequ("helpDirectory", lexptr))
	    {
	      yylval.symbol = HELP_DIR;
	      return HELP_DIR;
	    }
	  
	  if (!lexequ("helpFile", lexptr))
	    {
	      yylval.symbol = HELPFILE;
	      return HELPFILE;
	    }

	  if (!lexequ("horizontal", lexptr))
	    {
	      yylval.symbol = HORIZONTAL;
	      return HORIZONTAL;
	    }
	  
	  if (!lexequ("height", lexptr))
	    {
	      yylval.symbol = HEIGHT;
	      return HEIGHT;
	    }
	  
	  break;
	  
	case 'i':
	case 'I':

	  if (!lexequ("insensitive", lexptr))
	    {
	      yylval.symbol = INSENSITIVE;
	      return INSENSITIVE;
	    }

	  break;


	case 'l':
	case 'L':

	  if (!lexequ("layout", lexptr))
	    {
	      yylval.symbol = LAYOUT;
	      return LAYOUT;
	    }
	  
	  if (!lexequ("label", lexptr))
	    {
	      yylval.symbol = LABEL;
	      return LABEL;
	    }
	  
	  if (!lexequ("localmanaged", lexptr))
	    {
	      yylval.symbol = LOCALMANAGED;
	      return LOCALMANAGED;
	    }
	  
	  break;
	  
	case 'n':
	case 'N':

	  if (!lexequ("new", lexptr))
	    {
	      yylval.symbol = NEW;
	      return NEW;
	    }

	  break;

	case 'S':
	case 's':

	  if (!lexequ("searchClass", lexptr))
	    {
	      yylval.symbol = UFS_CLASS;
	      return UFS_CLASS;
	    }

	  if (!lexequ("searchRule", lexptr))
	    {
	      yylval.symbol = UFS_DEFAULTS;
	      return UFS_DEFAULTS;
	    }
	  
	  if (!lexequ("setaction", lexptr))
	    {
	      yylval.symbol = SETACTION;
	      return SETACTION;
	    }

	  if (!lexequ("setname", lexptr))
	    {
	      yylval.symbol = NAME;
	      return NAME;
	    }
	  
	  if (!lexequ("setbind", lexptr))
	    {
	      yylval.symbol = SETBIND;
	      return SETBIND;
	    }
	  
	  if (!lexequ("start", lexptr))
	    {
	      yylval.symbol = START;
	      return START;
	    }
	  
	  break;

	case 't':
	case 'T':
	  
	  if (!lexequ("targetClass", lexptr))
	    {
	      yylval.symbol = UFS_TARGET;
	      return UFS_TARGET;
	    }

	  break;

	case 'u':
	case 'U':
	  
	  if (!lexequ("ufnPath", lexptr))
	    {
	      yylval.symbol = UFN_PATH;
	      return UFN_PATH;
	    }

	  break;

	case 'v':
	case 'V':

	  if (!lexequ("vertical", lexptr))
	    {
	      yylval.symbol = VERTICAL;
	      return VERTICAL;
	    }
	  
	  break;
	  
	case 'w':
	case 'W':

	  if (!lexequ("width", lexptr))
	    {
	      yylval.symbol = WIDTH;
	      return WIDTH;
	    }
	  
	  break;

	default:
	  (void) strcpy(yylval.strval, lexeme);
	  return STRING;
	}

      (void) strcpy(yylval.strval, lexeme);
      return STRING;
    }
}

/* ARGSUSED */
int yyerror(err)
     char *err;
{
  (void) fprintf (stderr,
		  "\"%s\" Line %d.\n",
		  config_file_path,
		  config_line_num);

  (void) fprintf (stderr,
		  "Parse error at or near `%s'.\n",
		  lexeme);

  (void) fprintf(stderr, "Exiting.\n");
  
  exit(1);
}
