#ifndef _Xlookup_configure_h_
#define _Xlookup_configure_h_


#include <stdio.h>

#include "windows.h"
#include "defs.h"
#include "layout.h"
#include "actions.h"
#include "xdutil.h"

#include "attrt_list.h"
#include "sequence.h"

/*
 *	Used in configuration of window templates
 */

typedef enum 
{
  OT_action,
  OT_width,
  OT_height,
  OT_bitmap,
  OT_helpfile,
  OT_label,
  OT_name,
  OT_expand,
  OT_insensitive
} optionType;


typedef struct obj_res
{
  optionType type;
  union
    {
      String action_name;
      Dimension width;
      Dimension height;
      String bitmap_file;
      String help_file;
      String label;
      String name;
      Boolean expand;
      Boolean insensitive;
    } resource;

  struct obj_res *next;
} object_resource, *objectResource;

#define NULLObjRes (objectResource) NULL
#define resource_alloc() (objectResource) XtMalloc(sizeof(object_resource))

/*
 *	Structures for configuring directory operations
 */

/*
 *	This groups objectClasses into a generic entry type
 */
typedef struct _entry_class
{
  String name;

  AttributeType rdn_type;

  attrTypeList oc_list;
  attrTypeList default_attrs;

  struct _entry_class *next;
} entry_class, *entryClass;

#define NULLEntryClass (entryClass) NULL
#define entry_class_alloc() (entryClass) smalloc(sizeof(entry_class))


/*
 *	Attribute to objectClass rule
 */
typedef struct _attr_oc_rule
{
  AttributeType attr;
  AttributeType oc;

  struct _attr_oc_rule *next;
} attr_oc_rule, *attrOcRule;

#define NULLAttrOcRule (attrOcRule) NULL
#define attr_oc_alloc() (attrOcRule) smalloc(sizeof(attr_oc_rule))


/*
 *	Attribute value template
 */
typedef struct _attr_template
{
  AttributeType type;
  
  stringCell class_names;
  stringCell defaults;

  struct _attr_template *next;
} attr_template, *attrTemplate;

#define NULLAttrTemplate (attrTemplate) NULL
#define attr_tmpl_alloc() (attrTemplate) smalloc(sizeof(attr_template))


/*
 *    Attributes that are to be displayed
 */
typedef enum {ShowLongest, ShowAll} showValue;

typedef struct _read_attribute
{
  AttributeType type;
  showValue to_show;
  
  struct _read_attribute *next;
} read_attribute, *readAttribute;

#define NULLReadAttribute (readAttribute) NULL
#define read_attr_alloc() (readAttribute) XtMalloc(sizeof(read_attribute))


/*
 *   Friendly names for attributes
 */

typedef struct _attribute_name
{
  AttributeType type;
  String name;
  
  struct _attribute_name *next;
} attribute_name, *attributeName;

#define NULLAttributeName (attributeName) NULL
#define attr_name_alloc() (attributeName) XtMalloc(sizeof(attribute_name))


/*
 *	Procs
 */
void user_tailor(), parse_win_config();
void put_layout(), put_group();

Cardinal add_template_entry();
layoutMember add_prim_to_layout(), add_group_to_layout(), add_layout_branch(),
             add_branch_to_layout();

void add_to_startup_list();
void add_action();

translationList add_to_trans_list(), new_trans();
objectResource put_resource();

void set_entry_class(), set_base_class(), set_attr_oc_rule();
entryClass get_class_of_name();

void add_attribute_name();

void set_read_single_attrs(), set_read_all_attrs();
void add_display_attr();

void set_attr_template();
void set_search_rule();

void add_ufs_search_class();

void user_tailor();

#endif _Xlookup_configure_h_

