/*****************************************************************************

  layout.h

*****************************************************************************/

#ifndef _Xdir_layout_h
#define _Xdir_layout_h

#include "defs.h"
#include "windows.h"
#include "actions.h"

#define MAXTEMPLATES 255
#define MAXMEMBERS 64

typedef enum
{ /* orientation - of a layout group */
  horizontal,
  vertical
} orientation;

typedef enum
{
  REF_primitive,
  REF_layout
} reference_type;

typedef struct _prim_action
{
  actionID prim_action_id;
  struct _prim_action *next;
} prim_action, *primActionList;

#define prim_action_alloc() (primActionList) XtMalloc(sizeof(prim_action))
#define NULLPrimActionList (primActionList) NULL

typedef struct layout_member
{
  orientation orient;
  Cardinal template_id;

  reference_type type;
  union
    {
      Cardinal primitive_id;
      Cardinal layout_ref_id;
    } reference;

  struct layout_member *next_layout;

  /* Resources */
  Boolean expand;
  String bitmap_file;
  String help_file;
  String label;
  String name;

  /* Insensitivity rules */
#define I_DS_WAIT	0x01
#define I_NO_ERROR	0x02
#define I_NO_ENTRY	0x04
#define I_NO_LIST	0x08
#define I_NORMAL	0x10
  char insensitivity;

  /* Resources for primitives only */
  primActionList action_list;
  Dimension width;
  Dimension height;
} *layoutMember;

#define NULLLayout (layoutMember) 0
#define layout_alloc() (layoutMember) XtMalloc(sizeof(struct layout_member))

typedef struct window_template
{
  String template_name;
  Boolean parsed;
  file_org origin;
  layoutMember layout;
} *windowTemplate;

#define NULLTemplate (windowTemplate) 0
#define template_alloc()  (windowTemplate) \
                          XtMalloc(sizeof(struct window_template))

Widget make_from_layout();
layoutMember find_prim_member();
WidgetList find_prim_instances();
WidgetList find_named_instances();

#endif _Xdir_layout_h_


