/*
 * mua_selection.c
 *
 * To allow Inter Client Communication between and X.400 user agent (MUA) and 
 * and X.500 user agent (DUA) for the purpose of retrieving address 
 * information for the submission of messages.
 *
 * Overview:
 *
 * When the DUA comes up it takes ownership of the selection. If it loses the 
 * ownership (to another DUA) it takes no action. 
 * When the MUA requires an address it asks for the value of the selection 
 * with the target XA_STRING. The DUA will respond with four NULL separated 
 * strings - 
 * free form name - telephone number - DN - mhs oraddress
 *
 */

# include <X11/Intrinsic.h>

# include "mua_selection.h"

typedef int	(*IFP)();
# define NULL_IFP	((IFP)0)

# define NULLCP		((char*)0)
# define BUFSIZ		1024

#define STR_VAL(_s_)	(_s_ ? _s_ : "")
#define STR_LEN(_s_)	(_s_ ? strlen(_s_) : 0)



static	IFP 	DUAConvert = NULL_IFP;
static	IFP 	DUALose = NULL_IFP;
static caddr_t	DUAData = (caddr_t)0;
static  Atom	mua_atom;

Boolean	ConvertMUA();
void	LoseMUA();
void	DoneMUA();

OwnMUA(w, convert, lose, data)
Widget	w;
IFP	convert;
IFP	lose;
caddr_t	data;
{
    
    Display	*display = XtDisplay(w);

    DUAConvert = convert;
    DUALose = lose;
    DUAData = data;

    mua_atom = XInternAtom(display, MUA_ADDRESS, False);

    /* take ownership of the selection */

    XtOwnSelection(w, 
		   mua_atom,
		   XtLastTimestampProcessed(display),
		   ConvertMUA,
		   LoseMUA,
		   DoneMUA);
}
   
DisownMUA(w)
Widget	w;
{

    Display	*display = XtDisplay(w);

    XtDisownSelection(w, mua_atom, XtLastTimestampProcessed(display));

}
    
/* ARGSUSED */
Boolean
ConvertMUA(w, selection, target, type, value, length, format)
Widget		w;
Atom		*selection;
Atom		*target;
Atom		*type;
caddr_t		*value;
unsigned long	*length;
int		*format;
{
    
    Boolean	succeeded = False;
    Atom	*targets;
    char	*name = NULLCP, 
                *phone = NULLCP,
                *dn = NULLCP, 
                *or = NULLCP;
    static char buf[BUFSIZ];

    if(*target == XA_TARGETS(XtDisplay(w))) {
	
	/* ToDo: Look for default targets */

	*length = 1;
	*type = XA_ATOM;
	
	targets = (Atom*)XtMalloc(sizeof(Atom));

	*targets = XA_STRING;
	
	*value = (caddr_t)targets;

	succeeded = True;

    } else { 

	if((*target == XA_STRING) && DUAConvert) {

	    DUAConvert(DUAData, &name, &phone, &dn, &or);

	    if(or != NULLCP) {

		(void) sprintf(buf, "%s%c%s%c%s%c%s",
			       STR_VAL(name), NULL, STR_VAL(phone), NULL,
			       STR_VAL(dn), NULL, STR_VAL(or));
		
		*type = *target;
		*length = STR_LEN(name) + STR_LEN(phone) + 
		    STR_LEN(dn) + STR_LEN(or) + 3; /* 3 NULLs */
		*value = buf;
		*format = 8;

		succeeded = True;
	    }

	}

    }

    return(succeeded);
}

/* ARGSUSED */
void
LoseMUA(w, selection)
Widget	w;
Atom	*selection;
{

    if(DUALose) {
	(*DUALose)(DUAData);
    }
}


/* ARGSUSED */
void
DoneMUA(w, selection, target)
Widget	w;
Atom	*selection;
Atom	*target;
{

    /* we just want to maintain ownership of the selection value storage */
}
