#ifndef lint
static char *rcsid = "$Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/requests.c,v 0.95 91/09/02 11:37:27 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/requests.c,v 0.95 91/09/02 11:37:27 emsrdsm Exp Locker: emsrdsm $ */

/*****************************************************************************

  requests.c - Keep track of requests made and associated input id's.

*****************************************************************************/

#include <X11/Intrinsic.h>

#include "windows.h"
#include "requests.h"
#include "objects.h"
#include "callbacks.h"

#include "read_display.h"
#include "search_form.h"
#include "modify_form.h"
#include "modify_ops.h"

#include "query.h"

static void doQueryWork();

Cardinal request_count = 0;
assocRec request_list = NULLAssocRec;

extern XtAppContext app_context;

/*
 * - request_complete() -
 * The results from a request have been received. Delete it's record.
 *
 */
void request_complete(request_id)
     QCardinal request_id;
{
  assocRec this_request, prev_request, curr_request;
  int assoc_desc;

  for (this_request = request_list, prev_request = NULLAssocRec;
       this_request != NULLAssocRec && this_request->request_id != request_id;
       prev_request = this_request, this_request = this_request->next)
    ;

  /* This shouldn't happen */
  if (this_request == NULLAssocRec)
    return;

  if (prev_request != NULLAssocRec)
    prev_request->next = this_request->next;
  else
    request_list = this_request->next;

  for(assoc_desc = this_request->assoc_desc, curr_request = request_list;
      curr_request != NULLAssocRec && curr_request->assoc_desc != assoc_desc;
      curr_request = curr_request->next)
    ;

  /* If no other requests are using this association descriptor, remove
     the associated callback */
  if (curr_request == NULLAssocRec)
    XtRemoveInput(this_request->input_id);

  request_count--;
  XtFree((char *) this_request);
} /* request_complete */

/*
 * - request_invoked() -
 * A request has been invoked. If it's using a new association descriptor
 * then add an input callback.
 *
 */
void request_invoked(request_id,
		     type,
		     prim_widget,
		     object_id,
		     prim_id,
		     action_id,
		     trans_list)
     QCardinal request_id;
     request_type type;
     Widget prim_widget;
     Cardinal object_id;
     Cardinal prim_id;
     actionID action_id;
     translationList trans_list;
{
  XtInputId input_id;
  int assoc_desc;
  assocRec new_request, curr_request;

  /* Needs to be changed to get a list of ad's back, i.e. per task */
  assoc_desc = get_association_descriptor(request_id);

  /*
   * If this association descriptor is not already being used
   * then add another input callback to cover for it.
   *
   */
  for (curr_request = request_list;
       curr_request != NULLAssocRec && curr_request->assoc_desc != assoc_desc;
       curr_request = curr_request->next)
    ;
    
  if (curr_request == NULLAssocRec)
    input_id = XtAppAddInput(app_context,
			     assoc_desc,
			     XtInputReadMask,
			     doQueryWork,
			     (XtPointer) NULL);

  else input_id = curr_request->input_id;

  new_request = assoc_rec_alloc();
  new_request->input_id = input_id;
  new_request->assoc_desc = assoc_desc;
  new_request->request_id = request_id;
  new_request->type = type;

  new_request->this_widget = prim_widget;
  new_request->action_id = action_id;
  new_request->object_id = object_id;
  new_request->prim_id = prim_id;
  new_request->trans_list = trans_list;

  new_request->next = request_list;
  request_list = new_request;

  request_count++;
} /* request_invoked */

/*
 * - doQueryWork() -
 * Called when one of the dsap association descriptors is ready for input to
 * be read. 
 *
 */
/* ARGSUSED */
static void doQueryWork(clientdata, ad_ptr, id_ptr)
     XtPointer clientdata;
     int *ad_ptr;
     XtInputId *id_ptr;
{
  Cardinal request_id, object_id, prim_id;
  actionID action_id;
  translationList trans_list;
  Widget prim_widget;
  assocRec this_request;
  QCardinal *result_list = (QCardinal *) NULL,
            *report_list = (QCardinal *) NULL;
  Cardinal rqindex = 0, result_count, report_count;
  extern windowObject win_objects[];

  directory_wait(&result_list, &result_count, &report_list, &report_count);
  
  if (result_count == 0 && report_count == 0)
    return;

  for (rqindex = 0; rqindex < result_count; rqindex++)
    {
      request_id = result_list[rqindex];
      
      for (this_request = request_list;
	   this_request != NULLAssocRec
	   && this_request->request_id != request_id;
	   this_request = this_request->next)
	;
      
      /* This shouldn't happen */
      if (this_request == NULLAssocRec)
	continue;
      
      action_id = this_request->action_id;
      object_id = this_request->object_id;
      prim_widget = this_request->this_widget;
      trans_list = this_request->trans_list;
      prim_id = this_request->prim_id;
      
      switch (this_request->type)
	{
	case UFNAME:
	  gotMovetoResults(request_id, object_id, action_id, trans_list);
	  
	  break;
	  
	case UFSEARCH:
	  gotLookupResults(request_id, object_id, action_id, trans_list);
	  
	  break;

	case MODIFY_ENTRY:
	  gotSubmitModifyResult(request_id, object_id, action_id, trans_list);
	  
	  break;

	case DS_READ:
	  gotReadResult(request_id, object_id, action_id, trans_list);
	  
	  break;

	case ADD_ENTRY:
	  gotAddEntryResult(request_id, object_id, action_id, trans_list);

	  break;

	case DELETE_ENTRY:
	  gotDeleteEntryResult(request_id, object_id, action_id, trans_list);
	  
	  break;

	case ENTRY_TEMPLATE:
	  gotEntryTemplateResult(request_id, object_id, action_id, trans_list);

	  break;
	}
      
      doActionAux(prim_widget,
		  object_id,
		  prim_id,
		  action_id,
		  trans_list? trans_list->next : NULLTransList);
      
      request_complete(request_id);
    }


  for (rqindex = 0; rqindex < report_count; rqindex++)
    {
      request_id = report_list[rqindex];
      
      for (this_request = request_list;
	   this_request != NULLAssocRec
	   && this_request->request_id != request_id;
	   this_request = this_request->next)
	;
      
      /* This shouldn't happen */
      if (this_request == NULLAssocRec)
	continue;
      
      object_id = this_request->object_id;
      
      switch (this_request->type)
	{
	case UFNAME:
	  gotMovetoReport(request_id, object_id);
	  
	  break;

	case UFSEARCH:
	  gotLookupReport(request_id, object_id);
	  
	  break;
	}
    }


  doPrimitiveSensitivity(object_id);

  if (result_list != (QCardinal *) NULL)
    free((char *) result_list);

  if (report_list != (QCardinal *) NULL)
    free((char *) report_list);

  return;
} /* doQueryWork */







