#ifndef lint
static char *rcsid = "$Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/xutil.c,v 0.95 91/09/02 11:37:52 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/Pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/xdir/xutil.c,v 0.95 91/09/02 11:37:52 emsrdsm Exp Locker: emsrdsm $ */

/*******************************************************************

  xutil.c - Miscellaneous Xt related functions.

*******************************************************************/

#include <stdio.h>

#include "defs.h"
#include "windows.h"
#include "xdutil.h"
#include "xutil.h"

#include "quipu/common.h"

#include <X11/Xaw/Form.h>

/*
 *	Externals..
 */

extern Widget toplevel;
extern Display *dpy;


/*
 *	Local data
 */

static char err_msg[LINESIZE];


/*
 *	Local procedures..
 */

static void bitmap_format_warning();


/*
 * get_widget_geometry - get the position and dimensions of the given widget
 *			 using XtGetValues.
 */
void get_widget_geometry(widget, width_ptr, height_ptr, x_ptr, y_ptr)
     Widget widget;
     Dimension *width_ptr, *height_ptr;
     Position *x_ptr, *y_ptr;
{
  Arg args[MAXARGS];
  Cardinal count;
  
  if (widget == (Widget) 0)
    {
      if (width_ptr) *width_ptr = (Dimension) 0;
      if (height_ptr) *height_ptr = (Dimension) 0;

      if (x_ptr) *x_ptr = (Position) 0;
      if (y_ptr) *y_ptr = (Position) 0;

      return;
    }

  count = 0;
  if (width_ptr)
    {
      XtSetArg(args[count], XtNwidth, width_ptr); count++;
    }

  if (height_ptr)
    {
      XtSetArg(args[count], XtNheight, height_ptr); count++;
    }

  if (x_ptr)
    {
      XtSetArg(args[count], XtNx, x_ptr); count++;
    }

  if (y_ptr)
    {
      XtSetArg(args[count], XtNy, y_ptr); count++;
    }

  XtGetValues(widget, args, count);
} /* get_widget_geometry */

/*
 * set_widget_dimensions - Force geometry change. Can lead to weird results
 *                       if not used with care!
 */
void set_widget_dimensions(widget, width, height)
     Widget widget;
     Dimension width, height;
{
  Arg args[MAXARGS];
  Cardinal count;
  Dimension old_width, old_height, border_width;

  if (widget == NULLWidget) return;

  count = 0;
  if (width == 0)
    {
      XtSetArg(args[count], XtNwidth, &old_width); count++;
    }

  if (height == 0)
    {
      XtSetArg(args[count], XtNheight, &old_height); count++;
    }

  XtSetArg(args[count], XtNborderWidth, &border_width); count++;
  XtGetValues(widget, args, count);

  if (width == 0) width = old_width;
  if (height == 0) height = old_height;

  count = 0;

  XtSetArg(args[count], XtNwidth, width); count++;
  XtSetArg(args[count], XtNheight, height); count++;

  XtSetValues(widget, args, count);

  XtResizeWidget(widget, width, height, border_width);
} /* set_widget_dimensions */

/*
 * create_background_from_bitmap_file -
 *
 */
void create_background_from_bitmap_file(widget, bitmap_file_name)
     Widget widget;
     String bitmap_file_name;
{
  Arg args[MAXARGS];
  Cardinal count;
  Dimension width, height;
  Pixmap background;

  if ((background = create_bitmap_from_file(bitmap_file_name,
					    &width,
					    &height,
					    widget))
      == XtUnspecifiedPixmap)
    return;

  XawFormDoLayout(XtParent(widget), FALSE);

  count = 0;

  XtSetArg(args[count], XtNwidth, width); count++;
  XtSetArg(args[count], XtNheight, height); count++;

  XtSetArg(args[count], XtNbackgroundPixmap, background); count++;
  XtSetArg(args[count], XtNlabel, ""); count++;

  XtSetValues(widget, args, count);

  count = 0;

  XtSetArg(args[count], XtNresize, FALSE); count++;
  XtSetArg(args[count], XtNresizable, FALSE); count++;
  XtSetValues(widget, args, count);

  XawFormDoLayout(XtParent(widget), TRUE);
} /* create_background_from_bitmap_file */


/*
 * - create_bitmap_from_file() -
 *
 *
 */
Pixmap create_bitmap_from_file(bitmap_file_name, width_ptr, height_ptr, widget)
     String bitmap_file_name;
     Dimension *width_ptr, *height_ptr;
     Widget widget;
{
  FILE *bitmap_file;
  Dimension width, height;
  Cardinal bytes_wide, byte_count, count;
  char line[STRINGLEN];
  register String start, end;
  unsigned char *bytes, *byte_ptr;
  long temp_byte;
  Screen *screen;
  Pixmap return_pixmap;


  if ((bitmap_file = fopen(bitmap_file_name, "r")) == NULL)
    {
      (void) sprintf(err_msg, "Can't open bitmap file `%s'", bitmap_file_name);
      xluWarning(err_msg);

      (void) fclose(bitmap_file);

      return XtUnspecifiedPixmap;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;

      while (*end != '\0' && *end != '\n')
	end++;

      *end = '\0';
      
      start = end - 1;
      
      while (isdigit(*start))
	start--;

      start++;
      
      width = (Cardinal) atoi(start);
    }
  else
    {
      bitmap_format_warning(bitmap_file_name);

      (void) fclose(bitmap_file);
      
      return XtUnspecifiedPixmap;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;

      while (*end != '\0' && *end != '\n')
	end++;

      *end = '\0';
      

      start = end - 1;

      while (isdigit(*start))
	start--;

      start++;
      

      height = (Cardinal) atoi(start);
    }
  else
    {
      bitmap_format_warning(bitmap_file_name);
      
      (void) fclose(bitmap_file);

      return XtUnspecifiedPixmap;
    }
  
  bytes_wide = width / 8;

  if (width % 8 > 0)
    bytes_wide++;

  byte_count = bytes_wide * height;

  byte_ptr = bytes = (unsigned char *) XtMalloc(byte_count);
  
  while (1)
    {
      if (fgets(line, STRINGLEN, bitmap_file) == NULL)
	{
	  bitmap_format_warning(bitmap_file_name);
	  
	  (void) fclose(bitmap_file);
	  
	  XtFree((char *) bytes);

	  return XtUnspecifiedPixmap;
	}
      
      if (index(line, '{') != NULL) break;
    }

  count = 0;
  while (count < byte_count)
    {
      if (fgets(line, STRINGLEN, bitmap_file) == NULL)
	{
	  bitmap_format_warning(bitmap_file_name);
	  
	  (void) fclose(bitmap_file);

	  XtFree((char *) bytes);

	  return  XtUnspecifiedPixmap;
	}
    
      start = end = line;
      while (*end != '\0' && *end != '\n' && *end != ';')
	{
	  if ((start = index(start, 'x')) == NULL)
	    {
	      bitmap_format_warning(bitmap_file_name);

	      (void) fclose(bitmap_file);

	      XtFree((char *) bytes);

	      return XtUnspecifiedPixmap;
	    }
	  
	  start++;
	  end = start + 2;
	  
	  temp_byte = strtol(start, (char **) NULL, 16);

	  if (temp_byte < 0 || temp_byte > 255)
	    {
	      bitmap_format_warning(bitmap_file_name);

	      (void) fclose(bitmap_file);

	      XtFree((char *) bytes);

	      return XtUnspecifiedPixmap;
	    }

	  *byte_ptr++ = temp_byte;
	  count++;
	  end++;
	}
    }
  
  *width_ptr = width;
  *height_ptr = height;
  
  (void) fclose(bitmap_file);

  screen = XtScreen(toplevel);
  return_pixmap = XCreatePixmapFromBitmapData(dpy,
					      XtWindow(widget),
					      (char *) bytes,
					      width,
					      height,
					      BlackPixelOfScreen(screen),
					      WhitePixelOfScreen(screen),
					      DefaultDepthOfScreen(screen));

  XtFree((char *) bytes);

  return return_pixmap;
}


/*
 * get_dims_of_bitmap_file -
 *
 */
Boolean get_dims_of_bitmap_file(bitmap_file_name, width_ptr, height_ptr)
     String bitmap_file_name;
     int *width_ptr, *height_ptr;
{
  FILE *bitmap_file;
  Dimension width, height;
  char line[STRINGLEN];
  register String start, end;

  if ((bitmap_file = fopen(bitmap_file_name, "r")) == NULL)
    {
      (void) sprintf(err_msg,
		     "Cannot open bitmap file '%s'.\n", 
		     bitmap_file_name);

      xluWarning(err_msg);

      return FALSE;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;
      while (*end != '\0' && *end != '\n') end++;
      *end = '\0';
      
      start = end - 1;
      
      while (isdigit(*start)) start--;
      start++;

      width = (Cardinal) atoi(start);
    }
  else
    {
      bitmap_format_warning(bitmap_file_name);

      return FALSE;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;
      while (*end != '\0' && *end != '\n') end++;
      *end = '\0';
      
      start = end - 1;
      
      while (isdigit(*start)) start--;
      start++;
      
      height = (Cardinal) atoi(start);
    }
  else
    {
      return FALSE;
    }
  
  *width_ptr = (int) width;
  *height_ptr = (int) height;

  return TRUE;
}


static void bitmap_format_warning(filename)
     String filename;
{
  (void) sprintf(err_msg,
		 "Incorrect format in bitmap file `%s'.",
		 filename);
      
  xluWarning(err_msg);
}


/*
 * - get_text_dims() -
 * 
 * Calculate the width and height of a string using a given widgets' font.
 *
 */
void get_text_dims(string, widget, height_ptr, width_ptr)
     String string;
     Widget widget;
     int *height_ptr, *width_ptr;
{
  Cardinal count, linecount = 0;
  Arg args[MAXARGS];
  XFontStruct *font;
  int max_width, text_width;
  register String sptr, str;

  if (widget == NULLWidget)
    {
      *height_ptr = *width_ptr = 0;
      return;
    }

  count = 0;
  XtSetArg(args[count], XtNfont, &font); count++;
  XtGetValues(widget, args, count);

  if (string == NULLString || *string == '\0')
    {
      *height_ptr = FONTHEIGHT(font);
      *width_ptr = FONTWIDTH(font);

      return;
    }

  max_width = 0;
  sptr = str = string;

  while (str != NULLString && *str != '\0')
    {
      if (*str == '\n')
	{
	  linecount++;

	  *str = '\0';
	  
	  text_width = XTextWidth(font, sptr, strlen(sptr));

	  if (text_width > max_width) max_width = text_width;

	  *str = '\n';
	  sptr = ++str;
	}
      else
	str++;
    }

  if (sptr != NULLString)
    {
      text_width = XTextWidth(font, sptr, strlen(sptr));

      if (text_width > max_width) max_width = text_width;
    }

  linecount++;

  *width_ptr = max_width + FONTWIDTH(font);

  *height_ptr = linecount * FONTHEIGHT(font) + font->ascent;
} /* get_text_dims */


/*
 * - freeStorage() -
 * Frees memory associated with a widget when it is destroyed
 *
 */
/* ARGSUSED */
void freeStorage(w, closure, calldata)
     Widget w;
     XtPointer closure, calldata;
{
  XtFree((char *) closure);
}
